# This file was auto-generated by Fern from our API Definition.

import typing

import pydantic
import typing_extensions
from ..core.pydantic_utilities import IS_PYDANTIC_V2, UniversalBaseModel
from ..core.serialization import FieldMetadata
from .llm_as_judge_message_public import LlmAsJudgeMessagePublic
from .llm_as_judge_model_parameters_public import LlmAsJudgeModelParametersPublic
from .llm_as_judge_output_schema_public import LlmAsJudgeOutputSchemaPublic


class SpanLlmAsJudgeCodePublic(UniversalBaseModel):
    model: LlmAsJudgeModelParametersPublic
    messages: typing.List[LlmAsJudgeMessagePublic]
    variables: typing.Dict[str, str]
    schema_: typing_extensions.Annotated[typing.List[LlmAsJudgeOutputSchemaPublic], FieldMetadata(alias="schema")]

    if IS_PYDANTIC_V2:
        model_config: typing.ClassVar[pydantic.ConfigDict] = pydantic.ConfigDict(extra="allow", frozen=True)  # type: ignore # Pydantic v2
    else:

        class Config:
            frozen = True
            smart_union = True
            extra = pydantic.Extra.allow



import dataclasses
from typing import Any, Dict, Optional


@dataclasses.dataclass
class CloudRateLimit:
    remaining_limit: int
    remaining_limit_reset_time: int

    def retry_after(self) -> float:
        return self.remaining_limit_reset_time


def parse_rate_limit(rate_limit_response: Dict[str, Any]) -> Optional[CloudRateLimit]:
    rate_limit_reset_time = int(rate_limit_response.get("RateLimit-Reset", 0))
    if rate_limit_reset_time == 0:
        rate_limit_reset_time = int(rate_limit_response.get("ratelimit-reset", 0))

    if rate_limit_reset_time > 0:
        return CloudRateLimit(
            remaining_limit=0,
            remaining_limit_reset_time=rate_limit_reset_time,
        )

    return None



import logging
from typing import Any, Dict, List, Optional

from opik import llm_usage
from . import (
    openai_usage_extractor,
    anthropic_usage_extractor,
    google_generative_ai_usage_extractor,
    vertexai_usage_extractor,
    groq_usage_extractor,
    anthropic_vertexai_usage_extractor,
    bedrock_usage_extractor,
)
from . import provider_usage_extractor_protocol

LOGGER = logging.getLogger(__name__)

_REGISTERED_PROVIDER_USAGE_EXTRACTORS: List[
    provider_usage_extractor_protocol.ProviderUsageExtractorProtocol
] = [
    openai_usage_extractor.OpenAIUsageExtractor(),
    anthropic_usage_extractor.AnthropicUsageExtractor(),
    google_generative_ai_usage_extractor.GoogleGenerativeAIUsageExtractor(),
    vertexai_usage_extractor.VertexAIUsageExtractor(),
    groq_usage_extractor.GroqUsageExtractor(),
    anthropic_vertexai_usage_extractor.AnthropicVertexAIUsageExtractor(),
    bedrock_usage_extractor.BedrockUsageExtractor(),
]


def try_extract_provider_usage_data(
    run_dict: Dict[str, Any],
) -> Optional[llm_usage.LLMUsageInfo]:
    for extractor in _REGISTERED_PROVIDER_USAGE_EXTRACTORS:
        if not extractor.is_provider_run(run_dict):
            continue

        try:
            return extractor.get_llm_usage_info(run_dict)
        except Exception:
            LOGGER.warning(
                "Failed to extract usage data for presumably %s provider.",
                extractor.PROVIDER,
                exc_info=True,
            )
            return None

    return None



"""Experiment export functionality."""

import sys
from concurrent.futures import Future, ThreadPoolExecutor, as_completed
from datetime import datetime
from pathlib import Path
from typing import Optional, List, Dict, Tuple

import click
from rich.console import Console
from rich.progress import (
    Progress,
    SpinnerColumn,
    TextColumn,
    BarColumn,
    TaskProgressColumn,
)

import opik
from opik import exceptions
from .utils import (
    create_experiment_data_structure,
    debug_print,
    write_json_data,
    write_csv_data,
    print_export_summary,
    should_skip_file,
    trace_to_csv_rows,
)
from .dataset import export_experiment_datasets
from .prompt import (
    export_related_prompts_by_name,
    export_prompts_by_ids,
)

console = Console()

# Batch size for parallel trace fetching
BATCH_SIZE = 100
# Maximum number of concurrent workers for parallel execution
MAX_WORKERS = 20


def _fetch_trace_data(
    client: opik.Opik,
    trace_id: str,
    project_name_cache: dict[str, str],
    debug: bool,
) -> Optional[Tuple[str, dict, str]]:
    """Fetch trace and span data for a single trace ID.

    Returns:
        Tuple of (trace_id, trace_data_dict, project_name) or None if failed.
    """
    try:
        # Get trace by ID
        trace = client.get_trace_content(trace_id)

        # Get project name for this trace
        if not trace.project_id:
            return None

        # Get project name (use cache if available)
        if trace.project_id not in project_name_cache:
            try:
                project = client.get_project(trace.project_id)
                project_name_cache[trace.project_id] = project.name
            except Exception as e:
                if debug:
                    debug_print(
                        f"Warning: Could not get project for trace {trace_id}: {e}",
                        debug,
                    )
                return None

        project_name = project_name_cache[trace.project_id]

        # Get spans for this trace
        spans = client.search_spans(
            trace_id=trace_id,
            max_results=1000,
            truncate=False,
        )

        # Create trace data structure
        trace_data = {
            "trace": trace.model_dump(),
            "spans": [span.model_dump() for span in spans],
            "downloaded_at": datetime.now().isoformat(),
            "project_name": project_name,
        }

        return (trace_id, trace_data, project_name)
    except Exception as e:
        if debug:
            import traceback

            debug_print(
                f"Error fetching trace {trace_id}: {e}\n{traceback.format_exc()}", debug
            )
        return None


def _write_trace_file(
    trace_id: str,
    trace_data: dict,
    project_name: str,
    workspace_root: Path,
    format: str,
    force: bool,
    debug: bool,
) -> bool:
    """Write a single trace to file. Returns True if exported, False if skipped."""
    try:
        # Save trace in projects/PROJECT_NAME/ directory
        project_dir = workspace_root / "projects" / project_name
        project_dir.mkdir(parents=True, exist_ok=True)

        # Determine file path based on format
        if format.lower() == "csv":
            file_path = project_dir / f"trace_{trace_id}.csv"
        else:
            file_path = project_dir / f"trace_{trace_id}.json"

        # Check if file already exists and should be skipped
        if should_skip_file(file_path, force):
            if debug:
                debug_print(f"Skipping trace {trace_id} (already exists)", debug)
            return False

        # Save to file using the appropriate format
        if format.lower() == "csv":
            write_csv_data(trace_data, file_path, trace_to_csv_rows)
            if debug:
                debug_print(f"Wrote CSV file: {file_path}", debug)
        else:
            write_json_data(trace_data, file_path)
            if debug:
                debug_print(f"Wrote JSON file: {file_path}", debug)

        return True
    except Exception as e:
        console.print(f"[red]Error writing trace {trace_id} to file: {e}[/red]")
        if debug:
            import traceback

            debug_print(f"Traceback: {traceback.format_exc()}", debug)
        return False


def export_traces_by_ids(
    client: opik.Opik,
    trace_ids: List[str],
    workspace_root: Path,
    max_traces: Optional[int],
    format: str,
    debug: bool,
    force: bool,
) -> tuple[int, int]:
    """Export traces by their IDs using parallel batch processing.

    Traces are saved in projects/PROJECT_NAME/ directory based on each trace's project.
    Uses parallel execution to fetch traces/spans and write files concurrently.
    """
    exported_count = 0
    skipped_count = 0

    if max_traces:
        trace_ids = trace_ids[:max_traces]

    if not trace_ids:
        return 0, 0

    if debug:
        debug_print(
            f"Exporting {len(trace_ids)} trace(s) in batches of {BATCH_SIZE}", debug
        )

    # Cache project names to avoid repeated API calls (shared across threads)
    project_name_cache: dict[str, str] = {}

    # Use progress bar for trace export
    with Progress(
        SpinnerColumn(),
        TextColumn("[progress.description]{task.description}"),
        BarColumn(),
        TaskProgressColumn(),
        console=console,
    ) as progress:
        task = progress.add_task(
            f"Exporting {len(trace_ids)} traces...", total=len(trace_ids)
        )

        # Process traces in batches
        for batch_start in range(0, len(trace_ids), BATCH_SIZE):
            batch_end = min(batch_start + BATCH_SIZE, len(trace_ids))
            batch_trace_ids = trace_ids[batch_start:batch_end]

            if debug:
                debug_print(
                    f"Batch {batch_start // BATCH_SIZE + 1}: traces {batch_start + 1}-{batch_end}",
                    debug,
                )

            # Fetch trace data in parallel
            fetched_traces: dict[str, Tuple[dict, str]] = {}

            with ThreadPoolExecutor(max_workers=MAX_WORKERS) as fetch_executor:
                # Submit all trace fetch tasks and track trace_id for each future
                fetch_futures: Dict[Future[Optional[Tuple[str, dict, str]]], str] = {}
                for trace_id in batch_trace_ids:
                    fetch_future: Future[Optional[Tuple[str, dict, str]]] = (
                        fetch_executor.submit(
                            _fetch_trace_data,
                            client,
                            trace_id,
                            project_name_cache,
                            debug,
                        )
                    )
                    fetch_futures[fetch_future] = trace_id

                # Collect completed fetches
                for fetch_future in as_completed(fetch_futures):
                    trace_id = fetch_futures[fetch_future]
                    try:
                        result = fetch_future.result()
                        if result is not None:
                            fetched_trace_id, trace_data, project_name = result
                            fetched_traces[fetched_trace_id] = (
                                trace_data,
                                project_name,
                            )
                    except Exception as e:
                        if debug:
                            console.print(
                                f"[red]Error fetching trace {trace_id}: {e}[/red]"
                            )

            # Write files in parallel
            with ThreadPoolExecutor(max_workers=MAX_WORKERS) as write_executor:
                # Submit all write tasks and track trace_id for each future
                write_futures: Dict[Future[bool], str] = {}
                for trace_id, (trace_data, project_name) in fetched_traces.items():
                    write_future: Future[bool] = write_executor.submit(
                        _write_trace_file,
                        trace_id,
                        trace_data,
                        project_name,
                        workspace_root,
                        format,
                        force,
                        debug,
                    )
                    write_futures[write_future] = trace_id

                # Process completed writes
                for write_future in as_completed(write_futures):
                    trace_id = write_futures[write_future]
                    try:
                        if write_future.result():
                            exported_count += 1
                        else:
                            skipped_count += 1
                    except Exception as e:
                        if debug:
                            console.print(
                                f"[red]Error writing trace {trace_id}: {e}[/red]"
                            )
                    finally:
                        progress.update(
                            task,
                            advance=1,
                            description=f"Exported {exported_count}/{len(trace_ids)} traces",
                        )

                # Update progress for traces that failed to fetch
                for trace_id in batch_trace_ids:
                    if trace_id not in fetched_traces:
                        progress.update(task, advance=1)

    return exported_count, skipped_count


def export_experiment_by_id(
    client: opik.Opik,
    output_dir: Path,
    experiment_id: str,
    max_traces: Optional[int],
    force: bool,
    debug: bool,
    format: str,
    trace_ids_collector: Optional[set[str]] = None,
) -> tuple[Dict[str, int], int]:
    """Export a specific experiment by ID, including related datasets and traces.

    Returns:
        Tuple of (stats dictionary, file_written flag) where:
        - stats: Dictionary with keys "datasets", "prompts", "traces" and their counts
        - file_written: 1 if experiment file was written, 0 if skipped or error
    """
    try:
        console.print(f"[blue]Fetching experiment by ID: {experiment_id}[/blue]")

        # Get the specific experiment by ID
        experiment = client.get_experiment_by_id(experiment_id)
        if not experiment:
            console.print(f"[red]Experiment '{experiment_id}' not found[/red]")
            # Return empty stats and 0 for file written when not found
            return ({"datasets": 0, "prompts": 0, "traces": 0}, 0)

        debug_print(f"Found experiment: {experiment.name}", debug)

        # Get experiment items first (this can be slow for large experiments)
        console.print("[blue]Fetching experiment items...[/blue]")
        with Progress(
            SpinnerColumn(),
            TextColumn("[progress.description]{task.description}"),
            console=console,
        ) as progress:
            task = progress.add_task("Getting experiment items...", total=None)
            experiment_items = experiment.get_items()
            progress.update(task, description="Got experiment items")

        # Create experiment data structure
        experiment_data = create_experiment_data_structure(experiment, experiment_items)

        # Save experiment data
        # Include experiment ID in filename to handle multiple experiments with same name
        experiment_file = (
            output_dir / f"experiment_{experiment.name}_{experiment.id}.json"
        )
        file_already_exists = experiment_file.exists()
        experiment_file_written = False

        if not file_already_exists or force:
            write_json_data(experiment_data, experiment_file)
            experiment_file_written = True
            debug_print(
                f"Exported experiment: {experiment.name} (ID: {experiment.id})", debug
            )
        else:
            debug_print(
                f"Skipping experiment {experiment.name} (ID: {experiment.id}) (already exists)",
                debug,
            )

        # Related prompts and traces are handled at the batch level
        # Only export related prompts by name (this is experiment-specific and can't be easily deduplicated)
        stats = {
            "datasets": 0,
            "datasets_skipped": 0,
            "prompts": 0,
            "prompts_skipped": 0,
            "traces": 0,
            "traces_skipped": 0,
        }
        stats["prompts"] = export_related_prompts_by_name(
            client, experiment, output_dir, force, debug, format
        )

        # Collect trace IDs from experiment items (for batch export later)
        trace_ids = [item.trace_id for item in experiment_items if item.trace_id]
        if trace_ids_collector is not None:
            trace_ids_collector.update(trace_ids)

        # Traces are exported at batch level, so we don't export them here
        stats["traces"] = 0
        stats["traces_skipped"] = 0

        if debug:
            console.print(
                f"[green]Experiment {experiment.name} exported with stats: {stats}[/green]"
            )

        # Return stats dictionary and whether file was written
        return (stats, 1 if experiment_file_written else 0)

    except Exception as e:
        console.print(f"[red]Error exporting experiment {experiment_id}: {e}[/red]")
        # Return empty stats and 0 for file written on error
        return ({"datasets": 0, "prompts": 0, "traces": 0}, 0)


def export_experiment_by_name(
    name: str,
    workspace: str,
    output_path: str,
    dataset: Optional[str],
    max_traces: Optional[int],
    force: bool,
    debug: bool,
    format: str,
    api_key: Optional[str] = None,
) -> None:
    """Export an experiment by exact name."""
    try:
        if debug:
            debug_print(f"Exporting experiment: {name}", debug)

        # Initialize client
        if api_key:
            client = opik.Opik(api_key=api_key, workspace=workspace)
        else:
            client = opik.Opik(workspace=workspace)

        # Create output directory
        output_dir = Path(output_path) / workspace / "experiments"
        output_dir.mkdir(parents=True, exist_ok=True)
        datasets_dir = Path(output_path) / workspace / "datasets"
        datasets_dir.mkdir(parents=True, exist_ok=True)

        if debug:
            debug_print(f"Target directory: {output_dir}", debug)

        # Try to get experiments by exact name
        try:
            experiments = client.get_experiments_by_name(name)
            if not experiments:
                console.print(f"[red]Experiment '{name}' not found[/red]")
                return

            if debug:
                debug_print(
                    f"Found {len(experiments)} experiment(s) with name '{name}'", debug
                )

            if len(experiments) > 1:
                console.print(
                    f"[blue]Found {len(experiments)} experiments with name '{name}', exporting all of them[/blue]"
                )
        except Exception as e:
            console.print(f"[red]Experiment '{name}' not found: {e}[/red]")
            return

        # Filter experiments by dataset if specified (client-side filtering)
        if dataset:
            experiments = [exp for exp in experiments if exp.dataset_name == dataset]
            if not experiments:
                console.print(
                    f"[yellow]No experiments found with name '{name}' using dataset '{dataset}'[/yellow]"
                )
                return
            if debug:
                debug_print(
                    f"Filtered to {len(experiments)} experiment(s) using dataset '{dataset}'",
                    debug,
                )

        # Collect all unique resources from all experiments first
        unique_datasets = set()
        unique_prompt_ids: set[str] = set()

        # First pass: collect datasets and prompt IDs (these are available without fetching items)
        for experiment in experiments:
            if experiment.dataset_name:
                unique_datasets.add(experiment.dataset_name)
            # Get experiment data to access prompt_versions
            experiment_data = experiment.get_experiment_data()
            if experiment_data.prompt_versions:
                for prompt_version in experiment_data.prompt_versions:
                    if prompt_version.prompt_id:
                        unique_prompt_ids.add(prompt_version.prompt_id)

        # Export all unique datasets once before processing experiments
        datasets_exported = 0
        datasets_skipped = 0
        if unique_datasets:
            if len(unique_datasets) > 1:
                console.print(
                    f"[blue]Exporting {len(unique_datasets)} unique dataset(s) used by these experiments...[/blue]"
                )
            datasets_exported, datasets_skipped = export_experiment_datasets(
                client, unique_datasets, datasets_dir, format, debug, force
            )

        # Export all unique prompts once before processing experiments
        prompts_dir = output_dir.parent / "prompts"
        prompts_dir.mkdir(parents=True, exist_ok=True)
        prompts_exported = 0
        prompts_skipped = 0
        if unique_prompt_ids:
            if len(unique_prompt_ids) > 1:
                console.print(
                    f"[blue]Exporting {len(unique_prompt_ids)} unique prompt(s) used by these experiments...[/blue]"
                )
            prompts_exported, prompts_skipped = export_prompts_by_ids(
                client, unique_prompt_ids, prompts_dir, format, debug, force
            )

        # Collect all unique trace IDs from all experiments as we process them
        # We'll collect them during the first pass, then export once
        all_trace_ids: set[str] = set()

        # Export all matching experiments
        exported_count = 0
        skipped_count = 0

        # Aggregate stats from all experiments (prompts and traces already exported at batch level)
        aggregated_stats = {
            "prompts": 0,
            "prompts_skipped": 0,
        }

        for experiment in experiments:
            if debug:
                debug_print(
                    f"Exporting experiment: {experiment.name} (ID: {experiment.id})",
                    debug,
                )

            result = export_experiment_by_id(
                client,
                output_dir,
                experiment.id,
                max_traces,
                force,
                debug,
                format,
                all_trace_ids,
            )

            # result is a tuple: (stats_dict, file_written_flag)
            exp_stats, file_written = result
            # Aggregate stats (only related prompts, traces already handled)
            aggregated_stats["prompts"] += exp_stats.get("prompts", 0)
            aggregated_stats["prompts_skipped"] += exp_stats.get("prompts_skipped", 0)

            if file_written > 0:
                exported_count += 1
            else:
                skipped_count += 1

        # Export all unique traces once after collecting them from all experiments
        workspace_root = output_dir.parent
        traces_exported = 0
        traces_skipped = 0
        if all_trace_ids:
            trace_ids_list = list(all_trace_ids)
            if max_traces:
                trace_ids_list = trace_ids_list[:max_traces]
            if len(trace_ids_list) > 0:
                if len(all_trace_ids) > 1:
                    console.print(
                        f"[blue]Exporting {len(trace_ids_list)} unique trace(s) from these experiments...[/blue]"
                    )
                traces_exported, traces_skipped = export_traces_by_ids(
                    client, trace_ids_list, workspace_root, None, format, debug, force
                )

        # Collect statistics for summary
        stats = {
            "experiments": exported_count,
            "experiments_skipped": skipped_count,
            "datasets": datasets_exported,
            "datasets_skipped": datasets_skipped,
            "prompts": prompts_exported + aggregated_stats["prompts"],
            "prompts_skipped": prompts_skipped + aggregated_stats["prompts_skipped"],
            "traces": traces_exported,
            "traces_skipped": traces_skipped,
        }

        # Show export summary
        print_export_summary(stats, format)

        if exported_count > 0:
            if len(experiments) > 1:
                console.print(
                    f"[green]Successfully exported {exported_count} experiment(s) with name '{name}' to {output_dir}[/green]"
                )
            else:
                console.print(
                    f"[green]Successfully exported experiment '{name}' to {output_dir}[/green]"
                )
        else:
            console.print(
                f"[yellow]All {len(experiments)} experiment(s) with name '{name}' already exist (use --force to re-download)[/yellow]"
            )

    except Exception as e:
        console.print(f"[red]Error exporting experiment: {e}[/red]")
        sys.exit(1)


def export_experiment_by_name_or_id(
    name_or_id: str,
    workspace: str,
    output_path: str,
    dataset: Optional[str],
    max_traces: Optional[int],
    force: bool,
    debug: bool,
    format: str,
    api_key: Optional[str] = None,
) -> None:
    """Export an experiment by name or ID.

    First tries to get the experiment by ID. If not found, tries by name.
    """
    try:
        if debug:
            debug_print(f"Attempting to export experiment: {name_or_id}", debug)

        # Initialize client
        if api_key:
            client = opik.Opik(api_key=api_key, workspace=workspace)
        else:
            client = opik.Opik(workspace=workspace)

        # Create output directory
        output_dir = Path(output_path) / workspace / "experiments"
        output_dir.mkdir(parents=True, exist_ok=True)
        datasets_dir = Path(output_path) / workspace / "datasets"
        datasets_dir.mkdir(parents=True, exist_ok=True)

        # Try to get experiment by ID first
        try:
            if debug:
                debug_print(f"Trying to get experiment by ID: {name_or_id}", debug)
            experiment = client.get_experiment_by_id(name_or_id)

            # Successfully found by ID, export it
            if debug:
                debug_print(
                    f"Found experiment by ID: {experiment.name} (ID: {experiment.id})",
                    debug,
                )

            # Collect trace IDs as we export
            trace_ids_collector: set[str] = set()

            # Use the ID-based export function
            result = export_experiment_by_id(
                client,
                output_dir,
                name_or_id,
                max_traces,
                force,
                debug,
                format,
                trace_ids_collector,
            )

            exp_stats, file_written = result

            # Export related datasets
            unique_datasets = set()
            if experiment.dataset_name:
                unique_datasets.add(experiment.dataset_name)

            datasets_exported = 0
            datasets_skipped = 0
            if unique_datasets:
                datasets_exported, datasets_skipped = export_experiment_datasets(
                    client, unique_datasets, datasets_dir, format, debug, force
                )

            # Export traces collected from experiment items
            workspace_root = output_dir.parent
            traces_exported = 0
            traces_skipped = 0
            if trace_ids_collector:
                trace_ids_list = list(trace_ids_collector)
                if max_traces:
                    trace_ids_list = trace_ids_list[:max_traces]
                if len(trace_ids_list) > 0:
                    traces_exported, traces_skipped = export_traces_by_ids(
                        client,
                        trace_ids_list,
                        workspace_root,
                        None,
                        format,
                        debug,
                        force,
                    )

            # Collect statistics for summary
            stats = {
                "experiments": 1 if file_written > 0 else 0,
                "experiments_skipped": 0 if file_written > 0 else 1,
                "datasets": datasets_exported,
                "datasets_skipped": datasets_skipped,
                "prompts": exp_stats.get("prompts", 0),
                "prompts_skipped": exp_stats.get("prompts_skipped", 0),
                "traces": traces_exported,
                "traces_skipped": traces_skipped,
            }

            # Show export summary
            print_export_summary(stats, format)

            if file_written > 0:
                console.print(
                    f"[green]Successfully exported experiment '{experiment.name}' (ID: {experiment.id}) to {output_dir}[/green]"
                )
            else:
                console.print(
                    f"[yellow]Experiment '{experiment.name}' (ID: {experiment.id}) already exists (use --force to re-download)[/yellow]"
                )
            return

        except exceptions.ExperimentNotFound:
            # Not found by ID, try by name
            if debug:
                debug_print(
                    f"Experiment not found by ID, trying by name: {name_or_id}", debug
                )
            # Fall through to name-based export
            pass

        # Try by name (either because ID lookup failed or we're explicitly trying name)
        export_experiment_by_name(
            name_or_id,
            workspace,
            output_path,
            dataset,
            max_traces,
            force,
            debug,
            format,
            api_key,
        )

    except Exception as e:
        console.print(f"[red]Error exporting experiment: {e}[/red]")
        sys.exit(1)


@click.command(name="experiment")
@click.argument("name_or_id", type=str)
@click.option(
    "--dataset",
    type=str,
    help="Filter experiments by dataset name. Only experiments using this dataset will be exported.",
)
@click.option(
    "--max-traces",
    type=int,
    help="Maximum number of traces to export per experiment. Limits the total number of traces downloaded.",
)
@click.option(
    "--path",
    "-p",
    type=click.Path(file_okay=False, dir_okay=True, writable=True),
    default="opik_exports",
    help="Directory to save exported data. Defaults to opik_exports.",
)
@click.option(
    "--force",
    is_flag=True,
    help="Re-download items even if they already exist locally.",
)
@click.option(
    "--debug",
    is_flag=True,
    help="Enable debug output to show detailed information about the export process.",
)
@click.option(
    "--format",
    type=click.Choice(["json", "csv"], case_sensitive=False),
    default="json",
    help="Format for exporting data. Defaults to json.",
)
@click.pass_context
def export_experiment_command(
    ctx: click.Context,
    name_or_id: str,
    dataset: Optional[str],
    max_traces: Optional[int],
    path: str,
    force: bool,
    debug: bool,
    format: str,
) -> None:
    """Export an experiment by exact name to workspace/experiments.

    The command will first try to find the experiment by ID. If not found, it will try by name.
    """
    # Get workspace and API key from context
    workspace = ctx.obj["workspace"]
    api_key = ctx.obj.get("api_key") if ctx.obj else None
    export_experiment_by_name_or_id(
        name_or_id, workspace, path, dataset, max_traces, force, debug, format, api_key
    )



def generate_query(
    input: str,
    output: str,
) -> str:
    return f"""
        You are an impartial judge tasked with evaluating the quality and usefulness of AI-generated responses.
        Your evaluation should consider the following key factors:
        - Helpfulness: How well does it solve the user's problem?
        - Relevance: How well does it address the specific question?
        - Accuracy: Is the information correct and reliable?
        - Depth: Does it provide sufficient detail and explanation?
        - Creativity: Does it offer innovative or insightful perspectives when appropriate?
        - Level of detail: Is the amount of detail appropriate for the question?

        ###EVALUATION PROCESS###

        1. **ANALYZE** the user's question and the AI's response carefully
        2. **EVALUATE** how well the response meets each of the criteria above
        3. **CONSIDER** the overall effectiveness and usefulness of the response
        4. **PROVIDE** a clear, objective explanation for your evaluation
        5. **SCORE** the response on a scale from 0.0 to 1.0:
           - 1.0: Exceptional response that excels in all criteria
           - 0.8: Excellent response with minor room for improvement
           - 0.6: Good response that adequately addresses the question
           - 0.4: Fair response with significant room for improvement
           - 0.2: Poor response that barely addresses the question
           - 0.0: Completely inadequate or irrelevant response

        ###OUTPUT FORMAT###

        Your evaluation must be provided as a JSON object with exactly two fields:
        - "score": A float between 0.0 and 1.0
        - "reason": A brief, objective explanation justifying your score based on the criteria above

        Now, please evaluate the following:

        User Question: {input}
        AI Response: {output}

        Provide your evaluation in the specified JSON format.
        """



"""Common utilities for import functionality."""

from typing import Any, Dict, List, Optional

import opik
from opik.types import FeedbackScoreDict
from rich.console import Console
from rich.table import Table

console = Console()


def debug_print(message: str, debug: bool) -> None:
    """Print debug message only if debug is enabled."""
    if debug:
        console.print(f"[blue]{message}[/blue]")


def matches_name_pattern(name: str, pattern: Optional[str]) -> bool:
    """Check if a name matches the given pattern using case-insensitive substring matching."""
    if pattern is None:
        return True
    # Simple string matching - check if pattern is contained in name (case-insensitive)
    return pattern.lower() in name.lower()


def translate_trace_id(
    original_trace_id: str, trace_id_map: Dict[str, str]
) -> Optional[str]:
    """Translate an original trace id from export to the newly created id.

    Args:
        original_trace_id: The original trace ID from the export
        trace_id_map: Mapping from original trace IDs to new trace IDs (required, can be empty)

    Returns:
        The new trace ID if found in the map, None otherwise.
    """
    return trace_id_map.get(original_trace_id)


def find_dataset_item_by_content(
    dataset: opik.Dataset, expected_content: Dict[str, Any]
) -> Optional[str]:
    """Find a dataset item by matching its content.

    Compares all fields in expected_content (excluding 'id') with dataset items.
    """
    try:
        items = dataset.get_items()
        # Remove 'id' from expected_content for comparison
        expected_without_id = {k: v for k, v in expected_content.items() if k != "id"}

        for item in items:
            # Compare all fields (excluding 'id')
            item_without_id = {k: v for k, v in item.items() if k != "id"}
            if item_without_id == expected_without_id:
                return item.get("id")
    except Exception:
        pass
    return None


def create_dataset_item(dataset: opik.Dataset, item_data: Dict[str, Any]) -> str:
    """Create a dataset item and return its ID.

    Uses all fields from item_data (except 'id') to support flexible dataset schemas.
    """
    # Use all fields from item_data except 'id' (which is handled separately)
    new_item = {k: v for k, v in item_data.items() if k != "id" and v is not None}

    # Ensure item is not empty (backend requires non-empty data)
    if not new_item:
        raise ValueError(
            "Dataset item data is empty - at least one field must be provided"
        )

    dataset.insert([new_item])

    # Find the newly created item by matching all fields
    items = dataset.get_items()
    for item in items:
        # Compare all fields (excluding 'id')
        item_without_id = {k: v for k, v in item.items() if k != "id"}
        if item_without_id == new_item:
            item_id = item.get("id")
            if item_id is not None:
                return item_id

    dataset_name = getattr(dataset, "name", None)
    dataset_info = f", Dataset: {dataset_name!r}" if dataset_name else ""
    raise Exception(
        f"Failed to create dataset item. " f"Content: {new_item!r}{dataset_info}"
    )


def handle_trace_reference(item_data: Dict[str, Any]) -> Optional[str]:
    """Handle trace references from deduplicated exports."""
    trace_reference = item_data.get("trace_reference")
    if trace_reference:
        trace_id = trace_reference.get("trace_id")
        return trace_id

    # Fall back to direct trace_id
    return item_data.get("trace_id")


def clean_feedback_scores(
    feedback_scores: Optional[List[Dict[str, Any]]],
) -> Optional[List[FeedbackScoreDict]]:
    """Clean feedback scores by removing fields that are not allowed when creating them.

    Exported feedback scores include read-only fields like 'source', 'created_at', etc.
    that must be removed before importing.

    Allowed fields: name, value, category_name, reason
    """
    if not feedback_scores:
        return None

    cleaned_scores: List[FeedbackScoreDict] = []
    for score in feedback_scores:
        if not isinstance(score, dict):
            continue

        # Only keep allowed fields
        name = score.get("name")
        value = score.get("value")

        # Only add if name and value are present
        if not name or value is None:
            continue

        # Construct FeedbackScoreDict with required fields
        cleaned_score: FeedbackScoreDict = {
            "name": name,
            "value": value,
        }

        # Add optional fields if they exist
        if "category_name" in score:
            cleaned_score["category_name"] = score.get("category_name")
        if "reason" in score:
            cleaned_score["reason"] = score.get("reason")

        cleaned_scores.append(cleaned_score)

    return cleaned_scores if cleaned_scores else None


def print_import_summary(stats: Dict[str, int]) -> None:
    """Print a nice summary table of import statistics."""
    table = Table(
        title="📥 Import Summary", show_header=True, header_style="bold magenta"
    )
    table.add_column("Type", style="cyan", no_wrap=True)
    table.add_column("Imported", justify="right", style="green")
    table.add_column("Skipped", justify="right", style="yellow")
    table.add_column("Errors", justify="right", style="red")

    # Add rows for each type
    if (
        stats.get("experiments", 0) > 0
        or stats.get("experiments_skipped", 0) > 0
        or stats.get("experiments_errors", 0) > 0
    ):
        imported = stats.get("experiments", 0)
        skipped = stats.get("experiments_skipped", 0)
        errors = stats.get("experiments_errors", 0)
        table.add_row(
            "🧪 Experiments",
            str(imported),
            str(skipped) if skipped > 0 else "",
            str(errors) if errors > 0 else "",
        )

    if (
        stats.get("datasets", 0) > 0
        or stats.get("datasets_skipped", 0) > 0
        or stats.get("datasets_errors", 0) > 0
    ):
        imported = stats.get("datasets", 0)
        skipped = stats.get("datasets_skipped", 0)
        errors = stats.get("datasets_errors", 0)
        table.add_row(
            "📊 Datasets",
            str(imported),
            str(skipped) if skipped > 0 else "",
            str(errors) if errors > 0 else "",
        )

    if (
        stats.get("traces", 0) > 0
        or stats.get("traces_skipped", 0) > 0
        or stats.get("traces_errors", 0) > 0
    ):
        imported = stats.get("traces", 0)
        skipped = stats.get("traces_skipped", 0)
        errors = stats.get("traces_errors", 0)
        table.add_row(
            "🔍 Traces",
            str(imported),
            str(skipped) if skipped > 0 else "",
            str(errors) if errors > 0 else "",
        )

    if (
        stats.get("prompts", 0) > 0
        or stats.get("prompts_skipped", 0) > 0
        or stats.get("prompts_errors", 0) > 0
    ):
        imported = stats.get("prompts", 0)
        skipped = stats.get("prompts_skipped", 0)
        errors = stats.get("prompts_errors", 0)
        table.add_row(
            "💬 Prompts",
            str(imported),
            str(skipped) if skipped > 0 else "",
            str(errors) if errors > 0 else "",
        )

    if (
        stats.get("projects", 0) > 0
        or stats.get("projects_skipped", 0) > 0
        or stats.get("projects_errors", 0) > 0
    ):
        imported = stats.get("projects", 0)
        skipped = stats.get("projects_skipped", 0)
        errors = stats.get("projects_errors", 0)
        table.add_row(
            "📁 Projects",
            str(imported),
            str(skipped) if skipped > 0 else "",
            str(errors) if errors > 0 else "",
        )

    # Calculate totals
    total_imported = sum(
        [
            stats.get(key, 0)
            for key in ["experiments", "datasets", "traces", "prompts", "projects"]
        ]
    )
    total_skipped = sum(
        [
            stats.get(key, 0)
            for key in [
                "experiments_skipped",
                "datasets_skipped",
                "traces_skipped",
                "prompts_skipped",
                "projects_skipped",
            ]
        ]
    )
    total_errors = sum(
        [
            stats.get(key, 0)
            for key in [
                "experiments_errors",
                "datasets_errors",
                "traces_errors",
                "prompts_errors",
                "projects_errors",
            ]
        ]
    )

    table.add_row("", "", "", "", style="bold")
    table.add_row(
        "📦 Total",
        str(total_imported),
        str(total_skipped) if total_skipped > 0 else "",
        str(total_errors) if total_errors > 0 else "",
        style="bold green",
    )

    console.print()
    console.print(table)
    console.print()



import logging
from typing import Generator, Any, List, Optional, Callable, Dict
from opik.types import ErrorInfoDict

from opik.api_objects import trace, span
from opik.decorator import generator_wrappers, error_info_collector

from botocore import eventstream

LOGGER = logging.getLogger(__name__)


def wrap_stream(
    stream: eventstream.EventStream,
    capture_output: bool,
    span_to_end: span.SpanData,
    trace_to_end: Optional[trace.TraceData],
    generations_aggregator: Callable[[List[Any]], Any],
    response_metadata: Dict[str, Any],
    finally_callback: generator_wrappers.FinishGeneratorCallback,
) -> Generator[Any, None, None]:
    items: List[Dict[str, Any]] = []

    try:
        error_info: Optional[ErrorInfoDict] = None
        for item in stream:
            items.append(item)

            yield item
    except Exception as exception:
        LOGGER.debug(
            "Exception raised from botocore.eventstream.EventStream.",
            str(exception),
            exc_info=True,
        )
        error_info = error_info_collector.collect(exception)
        raise exception
    finally:
        if error_info is None:
            output = generations_aggregator(items)
            output["ResponseMetadata"] = response_metadata
        else:
            output = None

        finally_callback(
            output=output,
            error_info=error_info,
            generators_span_to_end=span_to_end,
            generators_trace_to_end=trace_to_end,
            capture_output=capture_output,
        )



import contextlib
from typing import Optional, Iterator

from opik.api_objects import experiment, opik_client, trace
from opik.api_objects.experiment import experiment_item
from opik.decorator import error_info_collector
from opik.types import ErrorInfoDict

import opik.context_storage as context_storage


@contextlib.contextmanager
def evaluate_llm_task_context(
    experiment: Optional[experiment.Experiment],
    dataset_item_id: str,
    trace_data: trace.TraceData,
    client: opik_client.Opik,
) -> Iterator[None]:
    error_info: Optional[ErrorInfoDict] = None
    try:
        context_storage.set_trace_data(trace_data)
        yield
    except Exception as exception:
        error_info = error_info_collector.collect(exception)
        raise
    finally:
        trace_data = context_storage.pop_trace_data()  # type: ignore

        assert trace_data is not None

        if error_info is not None:
            trace_data.error_info = error_info

        trace_data.init_end_time()

        client = client if client is not None else opik_client.get_client_cached()
        client.trace(**trace_data.as_parameters)

        # Only insert experiment item if an experiment is provided
        if experiment is not None:
            experiment_item_ = experiment_item.ExperimentItemReferences(
                dataset_item_id=dataset_item_id,
                trace_id=trace_data.id,
            )
            experiment.insert(experiment_items_references=[experiment_item_])


@contextlib.contextmanager
def evaluate_llm_task_result_spans_context(
    trace_data: trace.TraceData,
    client: opik_client.Opik,
) -> Iterator[None]:
    error_info: Optional[ErrorInfoDict] = None
    try:
        context_storage.set_trace_data(trace_data)
        yield
    except Exception as exception:
        error_info = error_info_collector.collect(exception)
        raise
    finally:
        trace_data = context_storage.pop_trace_data()  # type: ignore

        assert trace_data is not None

        if error_info is not None:
            trace_data.error_info = error_info

        trace_data.init_end_time()
        client.trace(**trace_data.as_parameters)



# This file was auto-generated by Fern from our API Definition.

import typing
from json.decoder import JSONDecodeError

from ..core.api_error import ApiError
from ..core.client_wrapper import AsyncClientWrapper, SyncClientWrapper
from ..core.http_response import AsyncHttpResponse, HttpResponse
from ..core.jsonable_encoder import jsonable_encoder
from ..core.pydantic_utilities import parse_obj_as
from ..core.request_options import RequestOptions
from ..errors.forbidden_error import ForbiddenError
from ..errors.not_found_error import NotFoundError
from ..errors.unauthorized_error import UnauthorizedError
from ..types.error_message import ErrorMessage
from ..types.provider_api_key_page_public import ProviderApiKeyPagePublic
from ..types.provider_api_key_public import ProviderApiKeyPublic
from .types.provider_api_key_write_provider import ProviderApiKeyWriteProvider

# this is used as the default value for optional parameters
OMIT = typing.cast(typing.Any, ...)


class RawLlmProviderKeyClient:
    def __init__(self, *, client_wrapper: SyncClientWrapper):
        self._client_wrapper = client_wrapper

    def delete_llm_provider_api_keys_batch(
        self, *, ids: typing.Sequence[str], request_options: typing.Optional[RequestOptions] = None
    ) -> HttpResponse[None]:
        """
        Delete LLM Provider's ApiKeys batch

        Parameters
        ----------
        ids : typing.Sequence[str]

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        HttpResponse[None]
        """
        _response = self._client_wrapper.httpx_client.request(
            "v1/private/llm-provider-key/delete",
            method="POST",
            json={
                "ids": ids,
            },
            headers={
                "content-type": "application/json",
            },
            request_options=request_options,
            omit=OMIT,
        )
        try:
            if 200 <= _response.status_code < 300:
                return HttpResponse(response=_response, data=None)
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response.text)
        raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response_json)

    def find_llm_provider_keys(
        self, *, request_options: typing.Optional[RequestOptions] = None
    ) -> HttpResponse[ProviderApiKeyPagePublic]:
        """
        Find LLM Provider's ApiKeys

        Parameters
        ----------
        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        HttpResponse[ProviderApiKeyPagePublic]
            LLMProviderApiKey resource
        """
        _response = self._client_wrapper.httpx_client.request(
            "v1/private/llm-provider-key",
            method="GET",
            request_options=request_options,
        )
        try:
            if 200 <= _response.status_code < 300:
                _data = typing.cast(
                    ProviderApiKeyPagePublic,
                    parse_obj_as(
                        type_=ProviderApiKeyPagePublic,  # type: ignore
                        object_=_response.json(),
                    ),
                )
                return HttpResponse(response=_response, data=_data)
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response.text)
        raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response_json)

    def store_llm_provider_api_key(
        self,
        *,
        provider: ProviderApiKeyWriteProvider,
        api_key: typing.Optional[str] = OMIT,
        name: typing.Optional[str] = OMIT,
        provider_name: typing.Optional[str] = OMIT,
        headers: typing.Optional[typing.Dict[str, str]] = OMIT,
        configuration: typing.Optional[typing.Dict[str, str]] = OMIT,
        base_url: typing.Optional[str] = OMIT,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> HttpResponse[None]:
        """
        Store LLM Provider's ApiKey

        Parameters
        ----------
        provider : ProviderApiKeyWriteProvider

        api_key : typing.Optional[str]

        name : typing.Optional[str]

        provider_name : typing.Optional[str]
            Provider name - required for custom LLM providers to uniquely identify them (e.g., 'ollama', 'vllm'). Must not be blank for custom providers. Should not be set for standard providers (OpenAI, Anthropic, etc.). This requirement is conditional and validation is enforced programmatically.

        headers : typing.Optional[typing.Dict[str, str]]

        configuration : typing.Optional[typing.Dict[str, str]]

        base_url : typing.Optional[str]

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        HttpResponse[None]
        """
        _response = self._client_wrapper.httpx_client.request(
            "v1/private/llm-provider-key",
            method="POST",
            json={
                "provider": provider,
                "api_key": api_key,
                "name": name,
                "provider_name": provider_name,
                "headers": headers,
                "configuration": configuration,
                "base_url": base_url,
            },
            headers={
                "content-type": "application/json",
            },
            request_options=request_options,
            omit=OMIT,
        )
        try:
            if 200 <= _response.status_code < 300:
                return HttpResponse(response=_response, data=None)
            if _response.status_code == 401:
                raise UnauthorizedError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Optional[typing.Any],
                        parse_obj_as(
                            type_=typing.Optional[typing.Any],  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 403:
                raise ForbiddenError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        ErrorMessage,
                        parse_obj_as(
                            type_=ErrorMessage,  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response.text)
        raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response_json)

    def get_llm_provider_api_key_by_id(
        self, id: str, *, request_options: typing.Optional[RequestOptions] = None
    ) -> HttpResponse[ProviderApiKeyPublic]:
        """
        Get LLM Provider's ApiKey by id

        Parameters
        ----------
        id : str

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        HttpResponse[ProviderApiKeyPublic]
            LLMProviderApiKey resource
        """
        _response = self._client_wrapper.httpx_client.request(
            f"v1/private/llm-provider-key/{jsonable_encoder(id)}",
            method="GET",
            request_options=request_options,
        )
        try:
            if 200 <= _response.status_code < 300:
                _data = typing.cast(
                    ProviderApiKeyPublic,
                    parse_obj_as(
                        type_=ProviderApiKeyPublic,  # type: ignore
                        object_=_response.json(),
                    ),
                )
                return HttpResponse(response=_response, data=_data)
            if _response.status_code == 404:
                raise NotFoundError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Optional[typing.Any],
                        parse_obj_as(
                            type_=typing.Optional[typing.Any],  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response.text)
        raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response_json)

    def update_llm_provider_api_key(
        self,
        id: str,
        *,
        api_key: typing.Optional[str] = OMIT,
        name: typing.Optional[str] = OMIT,
        provider_name: typing.Optional[str] = OMIT,
        headers: typing.Optional[typing.Dict[str, str]] = OMIT,
        configuration: typing.Optional[typing.Dict[str, str]] = OMIT,
        base_url: typing.Optional[str] = OMIT,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> HttpResponse[None]:
        """
        Update LLM Provider's ApiKey

        Parameters
        ----------
        id : str

        api_key : typing.Optional[str]

        name : typing.Optional[str]

        provider_name : typing.Optional[str]
            Provider name - can be set to migrate legacy custom LLM providers to the new multi-provider format. Once set, it cannot be changed. Should only be set for custom LLM providers.

        headers : typing.Optional[typing.Dict[str, str]]

        configuration : typing.Optional[typing.Dict[str, str]]

        base_url : typing.Optional[str]

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        HttpResponse[None]
        """
        _response = self._client_wrapper.httpx_client.request(
            f"v1/private/llm-provider-key/{jsonable_encoder(id)}",
            method="PATCH",
            json={
                "api_key": api_key,
                "name": name,
                "provider_name": provider_name,
                "headers": headers,
                "configuration": configuration,
                "base_url": base_url,
            },
            headers={
                "content-type": "application/json",
            },
            request_options=request_options,
            omit=OMIT,
        )
        try:
            if 200 <= _response.status_code < 300:
                return HttpResponse(response=_response, data=None)
            if _response.status_code == 401:
                raise UnauthorizedError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Optional[typing.Any],
                        parse_obj_as(
                            type_=typing.Optional[typing.Any],  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 403:
                raise ForbiddenError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        ErrorMessage,
                        parse_obj_as(
                            type_=ErrorMessage,  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 404:
                raise NotFoundError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Optional[typing.Any],
                        parse_obj_as(
                            type_=typing.Optional[typing.Any],  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response.text)
        raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response_json)


class AsyncRawLlmProviderKeyClient:
    def __init__(self, *, client_wrapper: AsyncClientWrapper):
        self._client_wrapper = client_wrapper

    async def delete_llm_provider_api_keys_batch(
        self, *, ids: typing.Sequence[str], request_options: typing.Optional[RequestOptions] = None
    ) -> AsyncHttpResponse[None]:
        """
        Delete LLM Provider's ApiKeys batch

        Parameters
        ----------
        ids : typing.Sequence[str]

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        AsyncHttpResponse[None]
        """
        _response = await self._client_wrapper.httpx_client.request(
            "v1/private/llm-provider-key/delete",
            method="POST",
            json={
                "ids": ids,
            },
            headers={
                "content-type": "application/json",
            },
            request_options=request_options,
            omit=OMIT,
        )
        try:
            if 200 <= _response.status_code < 300:
                return AsyncHttpResponse(response=_response, data=None)
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response.text)
        raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response_json)

    async def find_llm_provider_keys(
        self, *, request_options: typing.Optional[RequestOptions] = None
    ) -> AsyncHttpResponse[ProviderApiKeyPagePublic]:
        """
        Find LLM Provider's ApiKeys

        Parameters
        ----------
        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        AsyncHttpResponse[ProviderApiKeyPagePublic]
            LLMProviderApiKey resource
        """
        _response = await self._client_wrapper.httpx_client.request(
            "v1/private/llm-provider-key",
            method="GET",
            request_options=request_options,
        )
        try:
            if 200 <= _response.status_code < 300:
                _data = typing.cast(
                    ProviderApiKeyPagePublic,
                    parse_obj_as(
                        type_=ProviderApiKeyPagePublic,  # type: ignore
                        object_=_response.json(),
                    ),
                )
                return AsyncHttpResponse(response=_response, data=_data)
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response.text)
        raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response_json)

    async def store_llm_provider_api_key(
        self,
        *,
        provider: ProviderApiKeyWriteProvider,
        api_key: typing.Optional[str] = OMIT,
        name: typing.Optional[str] = OMIT,
        provider_name: typing.Optional[str] = OMIT,
        headers: typing.Optional[typing.Dict[str, str]] = OMIT,
        configuration: typing.Optional[typing.Dict[str, str]] = OMIT,
        base_url: typing.Optional[str] = OMIT,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> AsyncHttpResponse[None]:
        """
        Store LLM Provider's ApiKey

        Parameters
        ----------
        provider : ProviderApiKeyWriteProvider

        api_key : typing.Optional[str]

        name : typing.Optional[str]

        provider_name : typing.Optional[str]
            Provider name - required for custom LLM providers to uniquely identify them (e.g., 'ollama', 'vllm'). Must not be blank for custom providers. Should not be set for standard providers (OpenAI, Anthropic, etc.). This requirement is conditional and validation is enforced programmatically.

        headers : typing.Optional[typing.Dict[str, str]]

        configuration : typing.Optional[typing.Dict[str, str]]

        base_url : typing.Optional[str]

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        AsyncHttpResponse[None]
        """
        _response = await self._client_wrapper.httpx_client.request(
            "v1/private/llm-provider-key",
            method="POST",
            json={
                "provider": provider,
                "api_key": api_key,
                "name": name,
                "provider_name": provider_name,
                "headers": headers,
                "configuration": configuration,
                "base_url": base_url,
            },
            headers={
                "content-type": "application/json",
            },
            request_options=request_options,
            omit=OMIT,
        )
        try:
            if 200 <= _response.status_code < 300:
                return AsyncHttpResponse(response=_response, data=None)
            if _response.status_code == 401:
                raise UnauthorizedError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Optional[typing.Any],
                        parse_obj_as(
                            type_=typing.Optional[typing.Any],  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 403:
                raise ForbiddenError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        ErrorMessage,
                        parse_obj_as(
                            type_=ErrorMessage,  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response.text)
        raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response_json)

    async def get_llm_provider_api_key_by_id(
        self, id: str, *, request_options: typing.Optional[RequestOptions] = None
    ) -> AsyncHttpResponse[ProviderApiKeyPublic]:
        """
        Get LLM Provider's ApiKey by id

        Parameters
        ----------
        id : str

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        AsyncHttpResponse[ProviderApiKeyPublic]
            LLMProviderApiKey resource
        """
        _response = await self._client_wrapper.httpx_client.request(
            f"v1/private/llm-provider-key/{jsonable_encoder(id)}",
            method="GET",
            request_options=request_options,
        )
        try:
            if 200 <= _response.status_code < 300:
                _data = typing.cast(
                    ProviderApiKeyPublic,
                    parse_obj_as(
                        type_=ProviderApiKeyPublic,  # type: ignore
                        object_=_response.json(),
                    ),
                )
                return AsyncHttpResponse(response=_response, data=_data)
            if _response.status_code == 404:
                raise NotFoundError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Optional[typing.Any],
                        parse_obj_as(
                            type_=typing.Optional[typing.Any],  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response.text)
        raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response_json)

    async def update_llm_provider_api_key(
        self,
        id: str,
        *,
        api_key: typing.Optional[str] = OMIT,
        name: typing.Optional[str] = OMIT,
        provider_name: typing.Optional[str] = OMIT,
        headers: typing.Optional[typing.Dict[str, str]] = OMIT,
        configuration: typing.Optional[typing.Dict[str, str]] = OMIT,
        base_url: typing.Optional[str] = OMIT,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> AsyncHttpResponse[None]:
        """
        Update LLM Provider's ApiKey

        Parameters
        ----------
        id : str

        api_key : typing.Optional[str]

        name : typing.Optional[str]

        provider_name : typing.Optional[str]
            Provider name - can be set to migrate legacy custom LLM providers to the new multi-provider format. Once set, it cannot be changed. Should only be set for custom LLM providers.

        headers : typing.Optional[typing.Dict[str, str]]

        configuration : typing.Optional[typing.Dict[str, str]]

        base_url : typing.Optional[str]

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        AsyncHttpResponse[None]
        """
        _response = await self._client_wrapper.httpx_client.request(
            f"v1/private/llm-provider-key/{jsonable_encoder(id)}",
            method="PATCH",
            json={
                "api_key": api_key,
                "name": name,
                "provider_name": provider_name,
                "headers": headers,
                "configuration": configuration,
                "base_url": base_url,
            },
            headers={
                "content-type": "application/json",
            },
            request_options=request_options,
            omit=OMIT,
        )
        try:
            if 200 <= _response.status_code < 300:
                return AsyncHttpResponse(response=_response, data=None)
            if _response.status_code == 401:
                raise UnauthorizedError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Optional[typing.Any],
                        parse_obj_as(
                            type_=typing.Optional[typing.Any],  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 403:
                raise ForbiddenError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        ErrorMessage,
                        parse_obj_as(
                            type_=ErrorMessage,  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 404:
                raise NotFoundError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Optional[typing.Any],
                        parse_obj_as(
                            type_=typing.Optional[typing.Any],  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response.text)
        raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response_json)



from __future__ import annotations

from typing import Any, Callable, Optional, Sequence, Tuple, Union

from opik.exceptions import MetricComputationError
from opik.evaluation.metrics import base_metric, score_result

try:  # pragma: no cover - optional dependency
    from bert_score import score as bert_score_fn
except ImportError:  # pragma: no cover - optional dependency
    bert_score_fn = None


BertScoreFn = Callable[
    [Sequence[str], Union[Sequence[str], Sequence[Sequence[str]]]], Tuple[Any, Any, Any]
]


class BERTScore(base_metric.BaseMetric):
    """Wrapper around the `bert-score` library.

    Args:
        scorer_fn: Optional callable compatible with ``bert_score.score`` for
            dependency injection or advanced usage.
        model_type: Model checkpoint to use when loading the scorer. Ignored when
            ``scorer_fn`` is provided.
        lang: Two-letter language code used by the default scorer.
        rescale_with_baseline: Whether to rescale the score using the provided
            baseline statistics.
        device: Optional device string forwarded to ``bert_score`` (e.g., "cpu",
            "cuda").
    """

    def __init__(
        self,
        scorer_fn: Optional[BertScoreFn] = None,
        model_type: Optional[str] = "bert-base-uncased",
        lang: Optional[str] = "en",
        rescale_with_baseline: bool = False,
        device: Optional[str] = None,
        name: str = "bertscore_metric",
        track: bool = True,
        project_name: Optional[str] = None,
        **scorer_kwargs: Any,
    ) -> None:
        super().__init__(name=name, track=track, project_name=project_name)

        if scorer_fn is not None:
            self._scorer_fn = scorer_fn
        else:
            if bert_score_fn is None:  # pragma: no cover - optional dependency
                raise ImportError(
                    "BERTScore metric requires the optional 'bert-score' package. "
                    "Install via `pip install bert-score` or provide `scorer_fn`."
                )

            def _score(
                candidates: Sequence[str],
                references: Union[Sequence[str], Sequence[Sequence[str]]],
            ) -> Tuple[Any, Any, Any]:
                return bert_score_fn(
                    candidates,
                    references,
                    model_type=model_type,
                    lang=lang,
                    rescale_with_baseline=rescale_with_baseline,
                    device=device,
                    **scorer_kwargs,
                )

            self._scorer_fn = _score

    def score(
        self,
        output: str,
        reference: Union[str, Sequence[str], Sequence[Sequence[str]]],
        **ignored_kwargs: Any,
    ) -> score_result.ScoreResult:
        if not output.strip():
            raise MetricComputationError("Candidate is empty (BERTScore metric).")

        references: Union[Sequence[str], Sequence[Sequence[str]]]
        if isinstance(reference, str):
            references = [reference]
        else:
            references = reference
            if isinstance(reference, Sequence) and len(reference) == 0:
                raise MetricComputationError("Reference is empty (BERTScore metric).")

        precision, recall, f1 = self._scorer_fn([output], references)

        score_value = float(f1[0].item() if hasattr(f1[0], "item") else f1[0])
        metadata = {
            "precision": float(
                precision[0].item() if hasattr(precision[0], "item") else precision[0]
            ),
            "recall": float(
                recall[0].item() if hasattr(recall[0], "item") else recall[0]
            ),
        }

        return score_result.ScoreResult(
            value=score_value,
            name=self.name,
            reason=f"BERTScore F1: {score_value:.4f}",
            metadata=metadata,
        )



from typing import List, Any

from . import recursive_anonymizer, rules


class RulesAnonymizer(recursive_anonymizer.RecursiveAnonymizer):
    """An anonymizer that applies a list of rules sequentially to text data.

    This class takes a list of Rule objects and applies them to
    anonymize sensitive information in text.
    """

    def __init__(self, anonymizer_rules: List[rules.Rule], max_depth: int = 10):
        """Initialize the RulesAnonymizer with a list of rules.

        Args:
            anonymizer_rules: List of Rule objects to apply for anonymization.
            max_depth: Maximum recursion depth for nested data structures.
        """
        super().__init__(max_depth)
        self.rules = anonymizer_rules

    def anonymize_text(self, data: str, **kwargs: Any) -> str:
        """Apply all rules sequentially to the input text.

        Args:
            data: The text to anonymize.

        Returns:
            The anonymized text after applying all rules.
        """
        result = data
        for rule in self.rules:
            result = rule.apply(result)
        return result



from typing import Optional

import pydantic


class EvaluateConversationCoherenceResponse(pydantic.BaseModel):
    verdict: str
    reason: Optional[str] = pydantic.Field(default=None)

    __hash__ = object.__hash__


class ScoreReasonResponse(pydantic.BaseModel):
    reason: str

    __hash__ = object.__hash__



# This file was auto-generated by Fern from our API Definition.

import typing

import pydantic
from ..core.pydantic_utilities import IS_PYDANTIC_V2, UniversalBaseModel
from .audio_url import AudioUrl
from .image_url import ImageUrl
from .video_url import VideoUrl


class LlmAsJudgeMessageContent(UniversalBaseModel):
    type: str
    text: typing.Optional[str] = None
    image_url: typing.Optional[ImageUrl] = None
    video_url: typing.Optional[VideoUrl] = None
    audio_url: typing.Optional[AudioUrl] = None

    if IS_PYDANTIC_V2:
        model_config: typing.ClassVar[pydantic.ConfigDict] = pydantic.ConfigDict(extra="allow", frozen=True)  # type: ignore # Pydantic v2
    else:

        class Config:
            frozen = True
            smart_union = True
            extra = pydantic.Extra.allow



"""Project export functionality."""

import sys
from datetime import datetime
from pathlib import Path
from typing import Optional

import click
from rich.console import Console
from rich.progress import Progress, SpinnerColumn, TextColumn

import opik
from opik.rest_api.core.api_error import ApiError
from opik.rest_api.types.project_public import ProjectPublic
from .utils import (
    debug_print,
    matches_name_pattern,
    should_skip_file,
    print_export_summary,
    trace_to_csv_rows,
    write_csv_data,
    write_json_data,
)

console = Console()


def export_traces(
    client: opik.Opik,
    project_name: str,
    project_dir: Path,
    max_results: int,
    filter_string: Optional[str],
    project_name_filter: Optional[str] = None,
    format: str = "json",
    debug: bool = False,
    force: bool = False,
) -> tuple[int, int]:
    """Download traces and their spans with pagination support for large projects."""
    if debug:
        debug_print(
            f"DEBUG: _export_traces called with project_name: {project_name}, project_dir: {project_dir}",
            debug,
        )

    # Validate filter syntax before using it
    if filter_string:
        try:
            from opik.api_objects import opik_query_language

            # Try to parse the filter to validate syntax
            oql = opik_query_language.OpikQueryLanguage(filter_string)
            if oql.get_filter_expressions() is None and filter_string.strip():
                console.print(
                    f"[red]Error: Invalid filter syntax: Filter '{filter_string}' could not be parsed.[/red]"
                )
                console.print("[yellow]OQL filter syntax examples:[/yellow]")
                console.print('  status = "running"')
                console.print('  name contains "test"')
                console.print('  start_time >= "2024-01-01T00:00:00Z"')
                console.print("  usage.total_tokens > 1000")
                console.print(
                    "[yellow]Note: String values must be in double quotes, use = not :[/yellow]"
                )
                raise ValueError(
                    f"Invalid filter syntax: '{filter_string}' could not be parsed"
                )
        except ValueError as e:
            # If it's already our custom error message, re-raise it
            if "Invalid filter syntax" in str(e) and "could not be parsed" in str(e):
                raise
            # Otherwise, format the error nicely
            error_msg = str(e)
            console.print(f"[red]Error: Invalid filter syntax: {error_msg}[/red]")
            console.print("[yellow]OQL filter syntax examples:[/yellow]")
            console.print('  status = "running"')
            console.print('  name contains "test"')
            console.print('  start_time >= "2024-01-01T00:00:00Z"')
            console.print("  usage.total_tokens > 1000")
            console.print(
                "[yellow]Note: String values must be in double quotes, use = not :[/yellow]"
            )
            raise

    exported_count = 0
    skipped_count = 0
    page_size = min(100, max_results)  # Process in smaller batches
    current_page = 1
    total_processed = 0

    with Progress(
        SpinnerColumn(),
        TextColumn("[progress.description]{task.description}"),
        console=console,
    ) as progress:
        task = progress.add_task("Searching for traces...", total=None)

        while total_processed < max_results:
            # Calculate how many traces to fetch in this batch
            remaining = max_results - total_processed
            current_page_size = min(page_size, remaining)

            try:
                if debug:
                    debug_print(
                        f"DEBUG: Getting traces by project with project_name: {project_name}, filter: {filter_string}, page: {current_page}, size: {current_page_size}",
                        debug,
                    )

                # Parse filter to JSON format - backend expects JSON, not OQL string
                parsed_filter = None
                if filter_string:
                    from opik.api_objects import opik_query_language

                    oql = opik_query_language.OpikQueryLanguage(filter_string)
                    parsed_filter = oql.parsed_filters  # This is the JSON string
                    if debug:
                        debug_print(
                            f"DEBUG: Parsed filter to JSON: {parsed_filter}", debug
                        )

                # Use get_traces_by_project for better performance when we know the project name
                # Backend expects JSON string format for filters
                trace_page = client.rest_client.traces.get_traces_by_project(
                    project_name=project_name,
                    filters=parsed_filter,  # Pass parsed JSON string - backend expects JSON
                    page=current_page,
                    size=current_page_size,
                    truncate=False,  # Don't truncate data for download
                )
                traces = trace_page.content or []
                if debug:
                    debug_print(
                        f"DEBUG: Found {len(traces)} traces in project (page {current_page})",
                        debug,
                    )
            except Exception as e:
                # Check if it's an OQL parsing error and provide user-friendly message
                error_msg = str(e)
                if "Invalid value" in error_msg and (
                    "expected an string in double quotes" in error_msg
                    or "expected a string in double quotes" in error_msg
                ):
                    console.print(
                        "[red]Error: Invalid filter format in export query.[/red]"
                    )
                    console.print(
                        '[yellow]String values in filters must be in double quotes, e.g., status = "running"[/yellow]'
                    )
                    if debug:
                        debug_print(f"Technical details: {e}", debug)
                else:
                    console.print(f"[red]Error searching traces: {e}[/red]")
                break

            if not traces:
                # No more traces to process
                break

            # Update progress description to show current page
            progress.update(
                task,
                description=f"Downloading traces... (page {current_page}, {len(traces)} traces)",
            )

            # Filter traces by project name if specified
            if project_name_filter:
                original_count = len(traces)
                traces = [
                    trace
                    for trace in traces
                    if matches_name_pattern(trace.name or "", project_name_filter)
                ]
                if len(traces) < original_count:
                    console.print(
                        f"[blue]Filtered to {len(traces)} traces matching project '{project_name_filter}' in current batch[/blue]"
                    )

            if not traces:
                # No traces match the name pattern, but we might have more to process
                # Use original_traces for pagination, not the filtered empty list
                total_processed += current_page_size
                continue

            # Update progress for downloading
            progress.update(
                task,
                description=f"Downloading traces... (batch {total_processed // page_size + 1})",
            )

            # Download each trace with its spans
            for trace in traces:
                try:
                    # Get spans for this trace
                    spans = client.search_spans(
                        project_name=project_name,
                        trace_id=trace.id,
                        max_results=1000,  # Get all spans for the trace
                        truncate=False,
                    )

                    # Create trace data structure
                    trace_data = {
                        "trace": trace.model_dump(),
                        "spans": [span.model_dump() for span in spans],
                        "downloaded_at": datetime.now().isoformat(),
                        "project_name": project_name,
                    }

                    # Determine file path based on format
                    if format.lower() == "csv":
                        file_path = project_dir / f"trace_{trace.id}.csv"
                    else:
                        file_path = project_dir / f"trace_{trace.id}.json"

                    # Check if file already exists and should be skipped
                    if should_skip_file(file_path, force):
                        if debug:
                            debug_print(
                                f"Skipping trace {trace.id} (already exists)",
                                debug,
                            )
                        skipped_count += 1
                        total_processed += 1
                        continue

                    # Save to file using the appropriate format
                    try:
                        if format.lower() == "csv":
                            write_csv_data(trace_data, file_path, trace_to_csv_rows)
                            if debug:
                                debug_print(f"Wrote CSV file: {file_path}", debug)
                        else:
                            write_json_data(trace_data, file_path)
                            if debug:
                                debug_print(f"Wrote JSON file: {file_path}", debug)

                        exported_count += 1
                        total_processed += 1
                    except Exception as write_error:
                        console.print(
                            f"[red]Error writing trace {trace.id} to file: {write_error}[/red]"
                        )
                        if debug:
                            import traceback

                            debug_print(f"Traceback: {traceback.format_exc()}", debug)
                        continue

                except Exception as e:
                    console.print(f"[red]Error exporting trace {trace.id}: {e}[/red]")
                    continue

            # Update pagination for next iteration
            if traces:
                current_page += 1
            else:
                # No more traces to process
                break

            # If we got fewer traces than requested, we've reached the end
            if len(traces) < current_page_size:
                break

        # Final progress update
        if exported_count == 0:
            console.print("[yellow]No traces found in the project.[/yellow]")
        else:
            progress.update(task, description=f"Exported {exported_count} traces total")

    return exported_count, skipped_count


def export_project_by_name(
    name: str,
    workspace: str,
    output_path: str,
    filter_string: Optional[str],
    max_results: Optional[int],
    force: bool,
    debug: bool,
    format: str,
    api_key: Optional[str] = None,
) -> None:
    """Export a project by exact name."""
    try:
        if debug:
            debug_print(f"Exporting project: {name}", debug)

        # Validate filter syntax before doing anything else
        if filter_string:
            try:
                from opik.api_objects import opik_query_language

                # Try to parse the filter to validate syntax
                oql = opik_query_language.OpikQueryLanguage(filter_string)
                if oql.get_filter_expressions() is None and filter_string.strip():
                    console.print(
                        f"[red]Error: Invalid filter syntax: Filter '{filter_string}' could not be parsed.[/red]"
                    )
                    console.print("[yellow]OQL filter syntax examples:[/yellow]")
                    console.print('  status = "running"')
                    console.print('  name contains "test"')
                    console.print('  start_time >= "2024-01-01T00:00:00Z"')
                    console.print("  usage.total_tokens > 1000")
                    console.print(
                        "[yellow]Note: String values must be in double quotes, use = not :[/yellow]"
                    )
                    raise ValueError(
                        f"Invalid filter syntax: '{filter_string}' could not be parsed"
                    )
            except ValueError as e:
                # If it's already our custom error message, re-raise it
                if "Invalid filter syntax" in str(e) and "could not be parsed" in str(
                    e
                ):
                    raise
                # Otherwise, format the error nicely
                error_msg = str(e)
                console.print(f"[red]Error: Invalid filter syntax: {error_msg}[/red]")
                console.print("[yellow]OQL filter syntax examples:[/yellow]")
                console.print('  status = "running"')
                console.print('  name contains "test"')
                console.print('  start_time >= "2024-01-01T00:00:00Z"')
                console.print("  usage.total_tokens > 1000")
                console.print(
                    "[yellow]Note: String values must be in double quotes, use = not :[/yellow]"
                )
                raise

        # Initialize client
        if api_key:
            client = opik.Opik(api_key=api_key, workspace=workspace)
        else:
            client = opik.Opik(workspace=workspace)

        # Create output directory
        output_dir = Path(output_path) / workspace / "projects"
        output_dir.mkdir(parents=True, exist_ok=True)

        if debug:
            debug_print(f"Target directory: {output_dir}", debug)

        # Get projects and find exact match
        # Use name filter to narrow down results (backend does partial match, case-insensitive)
        # Then verify exact match on client side
        projects_response = client.rest_client.projects.find_projects(name=name)
        projects = projects_response.content or []
        matching_project = None

        for project in projects:
            if project.name == name:
                matching_project = project
                break

        # If not found with name filter, try paginating through all projects as fallback
        # This handles edge cases where the name filter might not work as expected
        if not matching_project:
            if debug:
                debug_print(
                    f"Project '{name}' not found in filtered results, searching all projects...",
                    debug,
                )
            # Paginate through all projects
            page = 1
            page_size = 100
            while True:
                projects_response = client.rest_client.projects.find_projects(
                    page=page, size=page_size
                )
                projects = projects_response.content or []
                if not projects:
                    break

                for project in projects:
                    if project.name == name:
                        matching_project = project
                        break

                if matching_project:
                    break

                # Check if there are more pages
                if projects_response.total is not None:
                    total_pages = (projects_response.total + page_size - 1) // page_size
                    if page >= total_pages:
                        break
                elif len(projects) < page_size:
                    # No more pages if we got fewer results than page size
                    break

                page += 1

        if not matching_project:
            console.print(f"[red]Project '{name}' not found[/red]")
            sys.exit(1)

        if debug:
            debug_print(f"Found project by exact match: {matching_project.name}", debug)

        # Export the project
        exported_count, traces_exported, traces_skipped = export_single_project(
            client,
            matching_project,
            output_dir,
            filter_string,
            max_results,
            force,
            debug,
            format,
        )

        # Collect statistics for summary using actual export counts
        stats = {
            "projects": 1 if exported_count > 0 else 0,
            "projects_skipped": 0 if exported_count > 0 else 1,
            "traces": traces_exported,
            "traces_skipped": traces_skipped,
        }

        # Show export summary
        print_export_summary(stats, format)

        if exported_count > 0:
            console.print(
                f"[green]Successfully exported project '{name}' to {output_dir}[/green]"
            )
        else:
            console.print(
                f"[yellow]Project '{name}' already exists (use --force to re-download)[/yellow]"
            )

    except ValueError as e:
        # Filter validation errors are already formatted, just exit
        if "Invalid filter syntax" in str(e):
            sys.exit(1)
        # Other ValueErrors should be shown
        console.print(f"[red]Error exporting project: {e}[/red]")
        sys.exit(1)
    except Exception as e:
        console.print(f"[red]Error exporting project: {e}[/red]")
        sys.exit(1)


def export_single_project(
    client: opik.Opik,
    project: ProjectPublic,
    output_dir: Path,
    filter_string: Optional[str],
    max_results: Optional[int],
    force: bool,
    debug: bool,
    format: str,
) -> tuple[int, int, int]:
    """Export a single project."""
    try:
        # Create project-specific directory for traces
        project_traces_dir = output_dir / project.name
        project_traces_dir.mkdir(parents=True, exist_ok=True)

        # Check if traces directory already has files in the requested format and force is not set
        # Skip early return if a filter is provided, as we need to check if existing traces match the filter
        if not force and not filter_string:
            # Only check for files in the requested format
            if format.lower() == "csv":
                existing_traces = list(project_traces_dir.glob("trace_*.csv"))
            else:
                existing_traces = list(project_traces_dir.glob("trace_*.json"))

            if existing_traces:
                if debug:
                    debug_print(
                        f"Skipping {project.name} (already has {len(existing_traces)} trace files in {format} format, use --force to re-download)",
                        debug,
                    )
                # Return project status, and trace counts (all skipped)
                return (1, 0, len(existing_traces))

        # Export related traces for this project
        traces_exported, traces_skipped = export_traces(
            client,
            project.name,
            project_traces_dir,
            max_results or 1000,  # Use provided max_results or default to 1000
            filter_string,
            None,  # project_name_filter
            format,
            debug,
            force,
        )

        # Project export only exports traces - datasets and prompts must be exported separately
        if traces_exported > 0:
            if debug:
                debug_print(
                    f"Exported project: {project.name} with {traces_exported} traces",
                    debug,
                )
            return (1, traces_exported, traces_skipped)
        elif traces_skipped > 0:
            # Traces were skipped (already exist)
            if debug:
                debug_print(
                    f"Project {project.name} already has {traces_skipped} trace files",
                    debug,
                )
            return (1, traces_exported, traces_skipped)
        else:
            # No traces found or exported
            if debug:
                debug_print(
                    f"No traces found in project: {project.name}",
                    debug,
                )
            # Still return 1 to indicate the project was processed
            # (the empty directory will remain, but that's expected if there are no traces)
            return (1, traces_exported, traces_skipped)

    except Exception as e:
        console.print(f"[red]Error exporting project {project.name}: {e}[/red]")
        return (0, 0, 0)


def export_project_by_name_or_id(
    name_or_id: str,
    workspace: str,
    output_path: str,
    filter_string: Optional[str],
    max_results: Optional[int],
    force: bool,
    debug: bool,
    format: str,
    api_key: Optional[str] = None,
) -> None:
    """Export a project by name or ID.

    First tries to get the project by ID. If not found, tries by name.
    """
    try:
        if debug:
            debug_print(f"Attempting to export project: {name_or_id}", debug)

        # Initialize client
        if api_key:
            client = opik.Opik(api_key=api_key, workspace=workspace)
        else:
            client = opik.Opik(workspace=workspace)

        # Create output directory
        output_dir = Path(output_path) / workspace / "projects"
        output_dir.mkdir(parents=True, exist_ok=True)

        if debug:
            debug_print(f"Target directory: {output_dir}", debug)

        # Try to get project by ID first
        try:
            if debug:
                debug_print(f"Trying to get project by ID: {name_or_id}", debug)
            project = client.get_project(name_or_id)

            # Successfully found by ID, export it
            if debug:
                debug_print(
                    f"Found project by ID: {project.name} (ID: {project.id})", debug
                )

            # Export the project
            exported_count, traces_exported, traces_skipped = export_single_project(
                client,
                project,
                output_dir,
                filter_string,
                max_results,
                force,
                debug,
                format,
            )

            # Show export summary
            stats = {
                "projects": exported_count,
                "traces": traces_exported,
                "traces_skipped": traces_skipped,
            }
            print_export_summary(stats, format)

            if exported_count > 0:
                console.print(
                    f"[green]Successfully exported project '{project.name}' (ID: {project.id}) to {output_dir}[/green]"
                )
            else:
                console.print(
                    f"[yellow]Project '{project.name}' (ID: {project.id}) already exists (use --force to re-download)[/yellow]"
                )
            return

        except ApiError as e:
            # Check if it's a 404 (not found) error
            if e.status_code == 404:
                # Not found by ID, try by name
                if debug:
                    debug_print(
                        f"Project not found by ID, trying by name: {name_or_id}", debug
                    )
                # Fall through to name-based export
                pass
            else:
                # Some other API error, re-raise it
                raise

        # Try by name (either because ID lookup failed or we're explicitly trying name)
        export_project_by_name(
            name_or_id,
            workspace,
            output_path,
            filter_string,
            max_results,
            force,
            debug,
            format,
            api_key,
        )

    except Exception as e:
        console.print(f"[red]Error exporting project: {e}[/red]")
        sys.exit(1)


@click.command(name="project")
@click.argument("name_or_id", type=str)
@click.option(
    "--filter",
    type=str,
    help="Filter string to narrow down traces using Opik Query Language (OQL).",
)
@click.option(
    "--max-results",
    type=int,
    help="Maximum number of traces to export. Limits the total number of traces downloaded.",
)
@click.option(
    "--path",
    "-p",
    type=click.Path(file_okay=False, dir_okay=True, writable=True),
    default="opik_exports",
    help="Directory to save exported data. Defaults to opik_exports.",
)
@click.option(
    "--force",
    is_flag=True,
    help="Re-download items even if they already exist locally.",
)
@click.option(
    "--debug",
    is_flag=True,
    help="Enable debug output to show detailed information about the export process.",
)
@click.option(
    "--format",
    type=click.Choice(["json", "csv"], case_sensitive=False),
    default="json",
    help="Format for exporting data. Defaults to json.",
)
@click.pass_context
def export_project_command(
    ctx: click.Context,
    name_or_id: str,
    filter: Optional[str],
    max_results: Optional[int],
    path: str,
    force: bool,
    debug: bool,
    format: str,
) -> None:
    """Export a project by name or ID to workspace/projects.

    The command will first try to find the project by ID. If not found, it will try by name.
    """
    # Get workspace and API key from context
    workspace = ctx.obj["workspace"]
    api_key = ctx.obj.get("api_key") if ctx.obj else None
    export_project_by_name_or_id(
        name_or_id, workspace, path, filter, max_results, force, debug, format, api_key
    )



"""
Harbor CLI integration with Opik tracking.

Usage:
    opik harbor run -d terminal-bench@head -a terminus_2 -m gpt-4.1
    opik harbor jobs start -c config.yaml
"""

import sys

import click


@click.command(
    name="harbor",
    context_settings={
        "ignore_unknown_options": True,
        "allow_extra_args": True,
        "allow_interspersed_args": False,
    },
)
@click.pass_context
def harbor(ctx: click.Context) -> None:
    """Run Harbor benchmarks with Opik tracking enabled."""
    try:
        import harbor  # noqa: F401
    except ImportError:
        raise click.ClickException(
            "Harbor is not installed. Install with: pip install harbor"
        )

    from opik.integrations.harbor import track_harbor

    track_harbor()

    from harbor.cli.main import app

    sys.argv = ["harbor"] + ctx.args
    app()



import enum
import logging
import sys

LOGGER = logging.getLogger(__name__)


def is_interactive() -> bool:
    """
    Determines if the current environment is interactive.

    Returns:
        bool: True if the environment is either running in a terminal,
              a Jupyter notebook, an IPython environment, or Google Colab.
              False otherwise.
    """
    return (
        sys.stdin.isatty()
        or _in_jupyter_environment()
        or _in_ipython_environment()
        or _in_colab_environment()
    )


def _in_jupyter_environment() -> bool:
    """
    Determine if the current environment is a Jupyter notebook.

    Returns:
        bool: True if running in a Jupyter notebook environment, otherwise False.
    """
    try:
        import IPython
    except Exception:
        return False

    ipy = IPython.get_ipython()
    if ipy is None or not hasattr(ipy, "kernel"):
        return False
    else:
        return True


def _in_ipython_environment() -> bool:
    """
    Determines if the current environment is an IPython environment.

    Returns:
        bool: True if the code is running in an IPython environment, False otherwise.
    """
    try:
        import IPython
    except Exception:
        return False

    ipy = IPython.get_ipython()
    if ipy is None:
        return False
    else:
        return True


def _in_colab_environment() -> bool:
    """
    Determines if the code is running within a Google Colab environment.

    Returns:
        bool: True if running in Google Colab, False otherwise.
    """
    try:
        import IPython
    except Exception:
        return False

    ipy = IPython.get_ipython()
    return "google.colab" in str(ipy)


def ask_user_for_approval(message: str) -> bool:
    """
    Prompt the user with a message for approval (Y/Yes/N/No).

    Args:
        message (str): The message to display to the user.

    Returns:
        bool: True if the user approves (Y/Yes/empty input), False if the user disapproves (N/No).

    Logs:
        Error when the user input is not recognized.
    """
    while True:
        users_choice = input(message).strip().upper()
        if users_choice in ("Y", "YES", ""):
            return True
        if users_choice in ("N", "NO"):
            return False
        LOGGER.error("Wrong choice. Please try again.")


class DeploymentType(enum.Enum):
    CLOUD = (1, "Opik Cloud (default)")
    SELF_HOSTED = (2, "Self-hosted Comet platform")
    LOCAL = (3, "Local deployment")

    @classmethod
    def find_by_value(cls, value: int) -> "DeploymentType":
        """
        Find the DeploymentType by its integer value.

        :param value: The integer value of the DeploymentType.
        :return: The corresponding DeploymentType.
        """
        for v in cls:
            if v.value[0] == value:
                return v
        raise ValueError(f"No DeploymentType with value '{value}'")


def ask_user_for_deployment_type() -> DeploymentType:
    """
    Asks the user to select a deployment type from the available Opik deployment options.
    Prompts the user until a valid selection is made.

    Returns:
        DeploymentType: The user's selected deployment type.
    """
    msg = ["Which Opik deployment do you want to log your traces to?"]

    for deployment in DeploymentType:
        msg.append(f"{deployment.value[0]} - {deployment.value[1]}")

    msg.append("\n> ")

    message_string = "\n".join(msg)

    while True:
        choice_str = input(message_string).strip()

        if choice_str not in ("1", "2", "3", ""):
            LOGGER.error("Wrong choice. Please try again.\n")
            continue

        if choice_str == "":
            choice_index = 1
        else:
            choice_index = int(choice_str)

        choice = DeploymentType.find_by_value(choice_index)

        return choice



from __future__ import annotations

from typing import Optional, Union

from opik.evaluation.metrics.llm_judges.g_eval import metric as g_eval_metric
from opik.evaluation.models import base_model


class AgentToolCorrectnessJudge(g_eval_metric.GEvalPreset):
    """
    Judge whether an agent invoked and interpreted tools correctly.

    Args:
        model: Optional model identifier or pre-configured ``OpikBaseModel``.
        track: Whether to automatically track judge outputs. Defaults to ``True``.
        project_name: Optional tracking project name.
        temperature: Sampling temperature supplied to the underlying model.

    Example:
        >>> from opik.evaluation.metrics import AgentToolCorrectnessJudge
        >>> judge = AgentToolCorrectnessJudge(model="gpt-4")
        >>> transcript = "Agent called search_tool and used the answer correctly."
        >>> result = judge.score(output=transcript)  # doctest: +SKIP
        >>> result.value  # doctest: +SKIP
        0.8
    """

    def __init__(
        self,
        model: Optional[Union[str, base_model.OpikBaseModel]] = None,
        track: bool = True,
        project_name: Optional[str] = None,
        temperature: float = 0.0,
    ) -> None:
        super().__init__(
            preset="agent_tool_correctness",
            model=model,
            track=track,
            project_name=project_name,
            temperature=temperature,
            name="agent_tool_correctness_judge",
        )


class AgentTaskCompletionJudge(g_eval_metric.GEvalPreset):
    """
    Evaluate whether an agent successfully completed the original task.

    Args:
        model: Optional model identifier or ``OpikBaseModel`` instance.
        track: Whether to automatically track judge outputs. Defaults to ``True``.
        project_name: Optional tracking project name.
        temperature: Sampling temperature for the underlying model.

    Example:
        >>> from opik.evaluation.metrics import AgentTaskCompletionJudge
        >>> judge = AgentTaskCompletionJudge(model="gpt-4")
        >>> result = judge.score(output="Agent delivered the requested summary.")  # doctest: +SKIP
        >>> result.value  # doctest: +SKIP
        0.9
    """

    def __init__(
        self,
        model: Optional[Union[str, base_model.OpikBaseModel]] = None,
        track: bool = True,
        project_name: Optional[str] = None,
        temperature: float = 0.0,
    ) -> None:
        super().__init__(
            preset="agent_task_completion",
            model=model,
            track=track,
            project_name=project_name,
            temperature=temperature,
            name="agent_task_completion_judge",
        )



"""LLM Juries: aggregate heterogeneous judges into a consensus score."""

from __future__ import annotations

from typing import Any, Dict, Iterable, List, Optional

from opik.evaluation.metrics.base_metric import BaseMetric
from opik.evaluation.metrics.score_result import ScoreResult
import opik.exceptions as exceptions


class LLMJuriesJudge(BaseMetric):
    """
    Aggregate multiple judge metrics into a consensus score.

    Args:
        judges: Iterable of judge metrics to execute for consensus.
        name: Display name for the aggregated result. Defaults to
            ``"llm_juries_judge"``.
        track: Whether to automatically track results. Defaults to ``True``.
        project_name: Optional tracking project name. Defaults to ``None``.

    Example:
        >>> from opik.evaluation.metrics import LLMJuriesJudge, ComplianceRiskJudge
        >>> juries = LLMJuriesJudge(judges=[ComplianceRiskJudge(model="gpt-4")])
        >>> result = juries.score(output="Financial guarantees provided.")  # doctest: +SKIP
        >>> result.value  # doctest: +SKIP
        0.6
    """

    def __init__(
        self,
        judges: Iterable[BaseMetric],
        name: str = "llm_juries_judge",
        track: bool = True,
        project_name: Optional[str] = None,
    ) -> None:
        super().__init__(name=name, track=track, project_name=project_name)
        self._judges = list(judges)
        if not self._judges:
            raise ValueError("LLMJuriesJudge requires at least one judge metric.")

    def score(self, *args: Any, **kwargs: Any) -> ScoreResult:
        precomputed: Optional[Dict[BaseMetric, ScoreResult]] = kwargs.pop(
            "precomputed", None
        )
        scores: List[ScoreResult] = []
        for judge in self._judges:
            if precomputed is not None and judge in precomputed:
                raw_result: Any = precomputed[judge]
            else:
                raw_result = judge.score(*args, **kwargs)
            judge_results = raw_result if isinstance(raw_result, list) else [raw_result]

            for result in judge_results:
                if not isinstance(result, ScoreResult):
                    raise exceptions.MetricComputationError(
                        f"Judge {judge.name} returned unexpected result type {type(result)!r}"
                    )
                if result.value < 0 or result.value > 1:
                    raise exceptions.MetricComputationError(
                        f"Judge {judge.name} returned out-of-range score {result.value}"
                    )
                scores.append(result)

        if not scores:
            raise exceptions.MetricComputationError("No judge scores produced")

        average = sum(res.value for res in scores) / len(scores)
        metadata = {
            "judge_scores": {res.name: res.value for res in scores},
        }
        reason = f"Averaged {len(scores)} judge scores"
        return ScoreResult(
            value=average, name=self.name, reason=reason, metadata=metadata
        )



from typing import List, Dict, Any, Optional
from . import guard
from .. import schemas
import functools


class PII(guard.Guard):
    """
    Guard that validates text for personally identifiable information (PII).
    """

    def __init__(
        self,
        blocked_entities: Optional[List[str]] = None,
        language: str = "en",
        threshold: float = 0.5,
    ) -> None:
        """
        Initialize a PII guard.

        Args:
            blocked_entities: List of PII entity types to block.
                Default entities include: "IP_ADDRESS", "PHONE_NUMBER", "PERSON", "MEDICAL_LICENSE", "URL", "EMAIL_ADDRESS", "IBAN_CODE".
            Supported entities can be found here: https://microsoft.github.io/presidio/supported_entities/
        """
        self._blocked_entities = blocked_entities
        self._language = language
        self._threshold = threshold

    @functools.lru_cache()
    def get_validation_configs(self) -> List[Dict[str, Any]]:
        """
        Get the validation configuration for PII detection.

        Returns:
            List containing the PII validation configuration
        """
        return [
            {
                "type": schemas.ValidationType.PII,
                "config": {
                    "entities": self._blocked_entities,
                    "language": self._language,
                    "threshold": self._threshold,
                },
            }
        ]



"""
Opik tracking integration for Harbor benchmark evaluation framework.

This module provides the `track_harbor` function to add Opik tracing to Harbor Jobs,
enabling real-time streaming of trial results to Opik for visualization and analysis.

Example:
    >>> from opik.integrations.harbor import track_harbor
    >>> from harbor.job import Job
    >>> import os
    >>>
    >>> os.environ["OPIK_PROJECT_NAME"] = "swebench-evaluation"
    >>>
    >>> job = Job(config)
    >>> tracked_job = track_harbor(job)
    >>> result = await tracked_job.run()

Or enable tracking globally (for CLI usage):
    >>> from opik.integrations.harbor import track_harbor
    >>> track_harbor()
    >>> # Now run Harbor code - it will be traced
"""

import functools
import logging
from typing import Any, Callable, Dict, List, Optional, Tuple
from typing_extensions import override

from harbor.job import Job
from harbor.models.trajectories.step import Step
from harbor.models.trial.result import TrialResult
from harbor.models.verifier.result import VerifierResult
from harbor.trial.trial import Trial
from harbor.verifier.verifier import Verifier

from opik import datetime_helpers, id_helpers, opik_context, track
from opik.api_objects import opik_client, span
from opik.decorator import arguments_helpers, base_track_decorator
from opik.types import FeedbackScoreDict, SpanType

from . import experiment_service

LOGGER = logging.getLogger(__name__)


class HarborTrialRunDecorator(base_track_decorator.BaseTrackDecorator):
    """
    Decorator for tracking Harbor Trial.run method.

    Sets the trace name based on trial configuration before the span/trace
    is sent to the backend.
    """

    @override
    def _start_span_inputs_preprocessor(
        self,
        func: Callable,
        track_options: arguments_helpers.TrackOptions,
        args: Tuple,
        kwargs: Dict[str, Any],
    ) -> arguments_helpers.StartSpanParameters:
        """Extract trial config and set trace name, input, metadata, and tags."""
        # Extract Trial instance from args (Trial.run is an instance method)
        if not args:
            # Fallback if no args (shouldn't happen for instance methods)
            name = (
                track_options.name if track_options.name is not None else func.__name__
            )
            return arguments_helpers.StartSpanParameters(
                name=name,
                input=None,
                type=track_options.type,
                tags=track_options.tags,
                metadata=track_options.metadata,
                project_name=track_options.project_name,
            )

        trial: Trial = args[0]
        config = trial.config

        # Build trace name from config
        trace_name = f"{config.agent.name}/{config.trial_name}"

        # Build input dict
        input_dict: Dict[str, Any] = {
            "trial_name": config.trial_name,
            "task": {
                "name": config.task.name
                if hasattr(config.task, "name")
                else str(config.task.path),
                "source": getattr(config.task, "source", None),
            },
            "agent": {
                "name": config.agent.name,
                "model": getattr(config.agent, "model_name", None),
            },
        }

        # Build metadata
        metadata = (
            track_options.metadata.copy() if track_options.metadata is not None else {}
        )
        metadata["created_from"] = "harbor"

        # Build tags
        tags = track_options.tags if track_options.tags is not None else []
        tags = list(tags)  # Make a copy to avoid mutating the original
        if "harbor" not in tags:
            tags.append("harbor")
        if config.agent.name not in tags:
            tags.append(config.agent.name)

        return arguments_helpers.StartSpanParameters(
            name=trace_name,
            input=input_dict,
            type=track_options.type,
            tags=tags,
            metadata=metadata,
            project_name=track_options.project_name,
        )

    @override
    def _end_span_inputs_preprocessor(
        self,
        output: Any,
        capture_output: bool,
        current_span_data: span.SpanData,
    ) -> arguments_helpers.EndSpanParameters:
        """Process output - minimal implementation since output is handled in _wrap_trial_run."""
        # Output is handled separately in _wrap_trial_run via opik_context.update_current_trace
        # So we don't need to process it here
        return arguments_helpers.EndSpanParameters(output=None)

    @override
    def _streams_handler(
        self,
        output: Any,
        capture_output: bool,
        generations_aggregator: Optional[Callable[[List[Any]], Any]],
    ) -> Optional[Any]:
        """No stream handling needed for Trial.run."""
        return None


def _rewards_to_feedback_scores(
    rewards: Optional[Dict[str, Any]],
    error: Optional[str] = None,
) -> List[FeedbackScoreDict]:
    """Convert Harbor verifier rewards to Opik feedback scores."""
    if rewards is None:
        return []

    feedback_scores: List[FeedbackScoreDict] = []
    for name, value in rewards.items():
        try:
            float_value = float(value)

            score = FeedbackScoreDict(name=name, value=float_value, reason=error)

            feedback_scores.append(score)
        except (ValueError, TypeError):
            LOGGER.warning(
                "Could not convert reward value to float: %s=%s", name, value
            )

    return feedback_scores


def _source_to_span_type(source: str) -> SpanType:
    """Convert ATIF step source to Opik span type."""
    if source == "agent":
        return "llm"
    return "general"


def _patch_step_class() -> None:
    """Patch the Harbor Step class to create Opik spans on instantiation."""
    # Check if already patched
    if hasattr(_patch_step_class, "_patched"):
        return

    original_init = Step.__init__

    @functools.wraps(original_init)
    def patched_init(self: Step, *args: Any, **kwargs: Any) -> None:
        original_init(self, *args, **kwargs)

        trace_data = opik_context.get_current_trace_data()
        if trace_data is None:
            return

        parent_span = opik_context.get_current_span_data()
        parent_span_id = parent_span.id if parent_span else None

        try:
            client = opik_client.get_client_cached()

            input_dict: Dict[str, Any] = {}
            if self.message:
                input_dict["message"] = self.message
            if self.tool_calls:
                input_dict["tool_calls"] = [
                    {
                        "tool_call_id": tc.tool_call_id,
                        "function_name": tc.function_name,
                        "arguments": tc.arguments,
                    }
                    for tc in self.tool_calls
                ]

            output_dict: Optional[Dict[str, Any]] = None
            if self.observation and self.observation.results:
                output_dict = {
                    "results": [
                        {"content": r.content} for r in self.observation.results
                    ]
                }

            metadata: Dict[str, Any] = {
                "source": self.source,
                "created_from": "harbor",
            }
            if self.reasoning_content:
                metadata["reasoning"] = self.reasoning_content

            usage: Optional[Dict[str, Any]] = None
            total_cost: Optional[float] = None
            if self.metrics:
                usage = {}
                if self.metrics.prompt_tokens is not None:
                    usage["prompt_tokens"] = self.metrics.prompt_tokens
                if self.metrics.completion_tokens is not None:
                    usage["completion_tokens"] = self.metrics.completion_tokens
                if self.metrics.prompt_tokens and self.metrics.completion_tokens:
                    usage["total_tokens"] = (
                        self.metrics.prompt_tokens + self.metrics.completion_tokens
                    )
                if not usage:
                    usage = None
                total_cost = getattr(self.metrics, "cost_usd", None)

            client.span(
                id=id_helpers.generate_id(),
                trace_id=trace_data.id,
                parent_span_id=parent_span_id,
                name=f"step_{self.step_id}",
                type=_source_to_span_type(self.source),
                start_time=datetime_helpers.parse_iso_timestamp(self.timestamp),
                input=input_dict if input_dict else None,
                output=output_dict,
                metadata=metadata,
                usage=usage,
                total_cost=total_cost,
                model=self.model_name if self.source == "agent" else None,
                tags=["harbor", self.source],
            )

        except Exception as e:
            LOGGER.debug("Failed to create span for step: %s", e)

    Step.__init__ = patched_init  # type: ignore
    setattr(_patch_step_class, "_patched", True)


def _enable_harbor_tracking(project_name: Optional[str] = None) -> None:
    """Internal: Enable Opik tracking for Harbor by patching classes.

    This patches Harbor's Trial and Verifier classes to add tracing.

    Args:
        project_name: Opik project name. If None, uses OPIK_PROJECT_NAME env var.
    """
    # Patch Trial methods (only if not already patched)
    if not hasattr(Trial.run, "opik_tracked"):
        Trial.run = _wrap_trial_run(Trial.run, project_name)

    if not hasattr(Trial._setup_environment, "opik_tracked"):
        Trial._setup_environment = _wrap_setup_environment(
            Trial._setup_environment, project_name
        )

    if not hasattr(Trial._setup_agent, "opik_tracked"):
        Trial._setup_agent = _wrap_setup_agent(Trial._setup_agent, project_name)

    if not hasattr(Trial._execute_agent, "opik_tracked"):
        Trial._execute_agent = _wrap_execute_agent(Trial._execute_agent, project_name)

    if not hasattr(Trial._run_verification, "opik_tracked"):
        Trial._run_verification = _wrap_run_verification(
            Trial._run_verification, project_name
        )

    # Patch Verifier (only if not already patched)
    if not hasattr(Verifier.verify, "opik_tracked"):
        Verifier.verify = _wrap_verify(Verifier.verify, project_name)

    # Patch Step class for real-time step tracking
    _patch_step_class()

    LOGGER.info("Opik tracking enabled for Harbor")


def track_harbor(
    job: Optional["Job"] = None,
    project_name: Optional[str] = None,
) -> Optional["Job"]:
    """Enable Opik tracking for Harbor.

    Can be called two ways:
    - track_harbor() - enables global tracking (for CLI usage)
    - track_harbor(job) - wraps a job and enables tracking (for SDK usage)

    Args:
        job: Optional Harbor Job instance. If provided, returns the same job.
        project_name: Opik project name. If None, uses OPIK_PROJECT_NAME env var.

    Returns:
        The job instance if provided, None otherwise.

    Example:
        >>> from opik.integrations.harbor import track_harbor
        >>> job = Job(config)
        >>> tracked_job = track_harbor(job)
        >>> result = await tracked_job.run()
    """
    _enable_harbor_tracking(project_name=project_name)
    return job


def _wrap_trial_run(original: Callable, project_name: Optional[str]) -> Callable:
    """Wrap Trial.run with tracing, feedback scores, and experiment linking."""

    decorator = HarborTrialRunDecorator()

    @decorator.track(
        tags=["harbor"],
        project_name=project_name,
        capture_output=False,
    )
    @functools.wraps(original)
    async def wrapped(self: Trial) -> TrialResult:
        config = self.config

        # Lazily setup experiment service if not already done
        # This ensures experiment tracking works for both SDK and CLI modes
        if experiment_service.get_service() is None:
            try:
                # Use job_id for consistent experiment naming
                experiment_name = (
                    f"harbor-job-{str(config.job_id)[:8]}" if config.job_id else None
                )
                # Build experiment config with agent/model info
                experiment_config: Dict[str, Any] = {
                    "agent_name": config.agent.name,
                }
                model_name = getattr(config.agent, "model_name", None)
                if model_name:
                    experiment_config["model_name"] = model_name

                LOGGER.debug(
                    "Lazily setting up experiment service: experiment_name=%s",
                    experiment_name,
                )
                experiment_service.setup_lazy(
                    experiment_name=experiment_name,
                    experiment_config=experiment_config,
                )
            except Exception as e:
                LOGGER.debug("Failed to lazily setup experiment service: %s", e)

        result: TrialResult = await original(self)

        # Update trace with output and feedback scores
        output_dict: Dict[str, Any] = {
            "trial_name": result.trial_name,
            "task_name": result.task_name,
        }
        if result.verifier_result and result.verifier_result.rewards:
            output_dict["rewards"] = result.verifier_result.rewards

        feedback_scores = None
        if result.verifier_result and result.verifier_result.rewards:
            # Get error message if available
            error_msg = getattr(result.verifier_result, "error", None) or getattr(
                result, "error", None
            )
            feedback_scores = _rewards_to_feedback_scores(
                result.verifier_result.rewards, error=error_msg
            )

        opik_context.update_current_trace(
            output=output_dict,
            feedback_scores=feedback_scores,
        )

        # Link to experiment
        trace_data = opik_context.get_current_trace_data()
        if trace_data is not None:
            service = experiment_service.get_service()
            LOGGER.debug(
                "Linking trial to experiment: trial=%s, trace_id=%s, service=%s",
                config.trial_name,
                trace_data.id,
                service,
            )
            if service is not None:
                source = getattr(config.task, "source", None)
                task_name = (
                    config.task.name
                    if hasattr(config.task, "name")
                    else str(config.task.path)
                )
                service.link_trial_to_experiment(
                    trial_name=config.trial_name,
                    trace_id=trace_data.id,
                    source=source,
                    task_name=task_name,
                )
            else:
                LOGGER.debug(
                    "No experiment service available, skipping experiment linking"
                )

        return result

    return wrapped


def _wrap_setup_environment(
    original: Callable, project_name: Optional[str]
) -> Callable:
    """Wrap Trial._setup_environment with tracing."""

    @track(name="setup_environment", tags=["harbor"], project_name=project_name)
    @functools.wraps(original)
    async def wrapped(self: Trial) -> None:
        opik_context.update_current_span(
            input={"phase": "environment_setup"},
            metadata={"created_from": "harbor"},
        )
        await original(self)
        opik_context.update_current_span(output={"status": "completed"})

    return wrapped


def _wrap_setup_agent(original: Callable, project_name: Optional[str]) -> Callable:
    """Wrap Trial._setup_agent with tracing."""

    @track(name="setup_agent", tags=["harbor"], project_name=project_name)
    @functools.wraps(original)
    async def wrapped(self: Trial) -> None:
        opik_context.update_current_span(
            input={"phase": "agent_setup"},
            metadata={"created_from": "harbor"},
        )
        await original(self)
        opik_context.update_current_span(output={"status": "completed"})

    return wrapped


def _wrap_execute_agent(original: Callable, project_name: Optional[str]) -> Callable:
    """Wrap Trial._execute_agent with tracing."""

    @track(name="execute_agent", tags=["harbor"], project_name=project_name)
    @functools.wraps(original)
    async def wrapped(self: Trial) -> None:
        input_dict = {}
        if hasattr(self, "_task") and self._task:
            input_dict["instruction"] = self._task.instruction
        opik_context.update_current_span(
            input=input_dict,
            metadata={"created_from": "harbor"},
        )
        await original(self)
        opik_context.update_current_span(output={"status": "completed"})

    return wrapped


def _wrap_run_verification(original: Callable, project_name: Optional[str]) -> Callable:
    """Wrap Trial._run_verification with tracing."""

    @track(name="run_verification", tags=["harbor"], project_name=project_name)
    @functools.wraps(original)
    async def wrapped(self: Trial) -> None:
        opik_context.update_current_span(
            input={"phase": "verification"},
            metadata={"created_from": "harbor"},
        )
        await original(self)
        opik_context.update_current_span(output={"status": "completed"})

    return wrapped


def _wrap_verify(original: Callable, project_name: Optional[str]) -> Callable:
    """Wrap Verifier.verify with tracing."""

    @track(name="verify", tags=["harbor"], project_name=project_name)
    @functools.wraps(original)
    async def wrapped(self: Verifier) -> VerifierResult:
        opik_context.update_current_span(
            input={"phase": "verify"},
            metadata={"created_from": "harbor"},
        )
        result: VerifierResult = await original(self)

        output_dict: Dict[str, Any] = {}
        if result.rewards:
            output_dict["rewards"] = result.rewards
        opik_context.update_current_span(
            output=output_dict if output_dict else {"status": "completed"}
        )

        return result

    return wrapped


def reset_harbor_tracking() -> None:
    """Reset Harbor tracking state for testing purposes.

    Resets the experiment service. Method patches remain active
    (they use `opik_tracked` to prevent double-patching).
    """
    experiment_service.reset()



from .api_classes import OpikArgs
from .helpers import (
    extract_opik_args,
    apply_opik_args_to_start_span_params,
    apply_opik_args_to_trace,
)

__all__ = [
    "OpikArgs",
    "extract_opik_args",
    "apply_opik_args_to_start_span_params",
    "apply_opik_args_to_trace",
]



import abc
from typing import Dict, Any, Optional

from . import base_metric, arguments_helpers
from .. import types as evaluation_types


class ScoreArgumentsValidator(abc.ABC):
    """Abstract base class for all metrics allowing validation of score method's arguments.
    It validates if arguments to be provided to the score method are valid, otherwise it raises an error.
    """

    @abc.abstractmethod
    def validate_score_arguments(
        self,
        score_kwargs: Dict[str, Any],
        key_mapping: Optional[evaluation_types.ScoringKeyMappingType],
    ) -> None:
        """The subclasses must implement this method to provide actual validation logic."""
        pass


def validate_score_arguments(
    metric: base_metric.BaseMetric,
    kwargs: Dict[str, Any],
    scoring_key_mapping: Optional[evaluation_types.ScoringKeyMappingType],
) -> None:
    if isinstance(metric, ScoreArgumentsValidator):
        metric.validate_score_arguments(
            score_kwargs=kwargs, key_mapping=scoring_key_mapping
        )
    else:
        arguments_helpers.raise_if_score_arguments_are_missing(
            score_function=metric.score,
            score_name=metric.name,
            kwargs=kwargs,
            scoring_key_mapping=scoring_key_mapping,
        )



"""Response types for Bedrock invoke_model operations."""

from dataclasses import dataclass
from typing import Any, Dict


@dataclass
class BedrockAggregatedResponse:
    """
    Response from invoke_model_with_response_stream after chunk aggregation.

    Contains the aggregated response in the provider's native format
    with standardized Bedrock usage format.
    """

    # Provider-native response structure
    native_response: Dict[str, Any]

    # Standardized Bedrock usage format
    usage: Dict[str, Any]  # {inputTokens, outputTokens, totalTokens}

    # Response metadata from Bedrock
    response_metadata: Dict[str, Any]

    def to_output_format(self) -> Dict[str, Any]:
        """Convert to output format for span logging."""
        return {"body": self.native_response}

    def to_metadata_format(self) -> Dict[str, Any]:
        """Convert to metadata format for span logging."""
        return {
            "created_from": "bedrock",
            **self.response_metadata,
        }



from typing import Union, cast, List

from opik.rest_api.types import span_write, trace_write
import opik.dict_utils as dict_utils

from . import base_batcher, sequence_splitter
from .. import messages, encoder_helpers


class CreateSpanMessageBatcher(base_batcher.BaseBatcher):
    def _create_batches_from_accumulated_messages(  # type: ignore
        self,
    ) -> List[messages.CreateSpansBatchMessage]:
        rest_spans: List[span_write.SpanWrite] = []

        for item in self._accumulated_messages:
            span_write_kwargs = item.as_payload_dict()
            cleaned_span_write_kwargs = dict_utils.remove_none_from_dict(
                span_write_kwargs
            )
            cleaned_span_write_kwargs = encoder_helpers.encode_and_anonymize(
                cleaned_span_write_kwargs,
                fields_to_anonymize=messages.CreateSpansBatchMessage.fields_to_anonymize(),
                object_type="span",
            )
            rest_spans.append(span_write.SpanWrite(**cleaned_span_write_kwargs))

        memory_limited_batches = sequence_splitter.split_into_batches(
            items=rest_spans,
            max_payload_size_MB=self._batch_memory_limit_mb,
        )

        batches = []
        for batch in memory_limited_batches:
            batches.append(messages.CreateSpansBatchMessage(batch=batch))

        return batches

    def add(self, message: messages.CreateSpanMessage) -> None:  # type: ignore
        # remove any duplicate start span message from the batch that was already added
        if message.end_time is not None:
            self._remove_matching_messages(lambda x: x.span_id == message.span_id)  # type: ignore

        return super().add(message)


class CreateTraceMessageBatcher(base_batcher.BaseBatcher):
    def _create_batches_from_accumulated_messages(  # type: ignore
        self,
    ) -> List[messages.CreateTraceBatchMessage]:
        rest_traces: List[trace_write.TraceWrite] = []

        for item in self._accumulated_messages:
            trace_write_kwargs = item.as_payload_dict()
            cleaned_trace_write_kwargs = dict_utils.remove_none_from_dict(
                trace_write_kwargs
            )
            cleaned_trace_write_kwargs = encoder_helpers.encode_and_anonymize(
                cleaned_trace_write_kwargs,
                fields_to_anonymize=messages.CreateTraceBatchMessage.fields_to_anonymize(),
                object_type="trace",
            )
            rest_traces.append(trace_write.TraceWrite(**cleaned_trace_write_kwargs))

        memory_limited_batches = sequence_splitter.split_into_batches(
            items=rest_traces,
            max_payload_size_MB=self._batch_memory_limit_mb,
        )

        batches = []
        for batch in memory_limited_batches:
            batches.append(messages.CreateTraceBatchMessage(batch=batch))

        return batches

    def add(self, message: messages.CreateTraceMessage) -> None:  # type: ignore
        # remove any duplicate start trace message from the batch that was already added
        if message.end_time is not None:
            self._remove_matching_messages(lambda x: x.trace_id == message.trace_id)  # type: ignore

        return super().add(message)


class BaseAddFeedbackScoresBatchMessageBatcher(base_batcher.BaseBatcher):
    def _create_batches_from_accumulated_messages(  # type: ignore
        self,
    ) -> List[
        Union[
            messages.AddSpanFeedbackScoresBatchMessage,
            messages.AddTraceFeedbackScoresBatchMessage,
            messages.AddThreadsFeedbackScoresBatchMessage,
        ]
    ]:
        return super()._create_batches_from_accumulated_messages()  # type: ignore

    def add(  # type: ignore
        self,
        message: Union[
            messages.AddSpanFeedbackScoresBatchMessage,
            messages.AddTraceFeedbackScoresBatchMessage,
            messages.AddThreadsFeedbackScoresBatchMessage,
        ],
    ) -> None:
        new_messages = message.batch
        n_new_messages = len(new_messages)
        n_accumulated_messages = len(self._accumulated_messages)

        if n_new_messages + n_accumulated_messages >= self._max_batch_size:
            free_space_in_accumulator = self._max_batch_size - n_accumulated_messages

            messages_that_fit_in_batch = new_messages[:free_space_in_accumulator]
            messages_that_dont_fit_in_batch = new_messages[free_space_in_accumulator:]

            self._accumulated_messages += messages_that_fit_in_batch
            new_messages = messages_that_dont_fit_in_batch
            self.flush()

        self._accumulated_messages += new_messages


class AddSpanFeedbackScoresBatchMessageBatcher(
    BaseAddFeedbackScoresBatchMessageBatcher
):
    def _create_batches_from_accumulated_messages(  # type: ignore
        self,
    ) -> List[messages.AddSpanFeedbackScoresBatchMessage]:
        return [
            messages.AddSpanFeedbackScoresBatchMessage(
                batch=self._accumulated_messages,  # type: ignore
                supports_batching=False,
            )
        ]


class AddTraceFeedbackScoresBatchMessageBatcher(
    BaseAddFeedbackScoresBatchMessageBatcher
):
    def _create_batches_from_accumulated_messages(  # type: ignore
        self,
    ) -> List[messages.AddTraceFeedbackScoresBatchMessage]:
        return [
            messages.AddTraceFeedbackScoresBatchMessage(
                batch=self._accumulated_messages,  # type: ignore
                supports_batching=False,
            )
        ]


class AddThreadsFeedbackScoresBatchMessageBatcher(
    BaseAddFeedbackScoresBatchMessageBatcher
):
    def _create_batches_from_accumulated_messages(  # type: ignore
        self,
    ) -> List[messages.AddThreadsFeedbackScoresBatchMessage]:
        return [
            messages.AddThreadsFeedbackScoresBatchMessage(
                batch=self._accumulated_messages,  # type: ignore
                supports_batching=False,
            )
        ]


class GuardrailBatchMessageBatcher(base_batcher.BaseBatcher):
    def _create_batches_from_accumulated_messages(  # type: ignore
        self,
    ) -> List[messages.GuardrailBatchMessage]:
        batch = []

        for batch_message in self._accumulated_messages:
            batch_of_message_item = cast(
                messages.GuardrailBatchMessage, batch_message
            ).batch
            batch.extend(batch_of_message_item)

        return [messages.GuardrailBatchMessage(batch=batch, supports_batching=False)]  # type: ignore

    def add(self, message: messages.GuardrailBatchMessage) -> None:  # type: ignore
        return super().add(message)


class CreateExperimentItemsBatchMessageBatcher(base_batcher.BaseBatcher):
    def _create_batches_from_accumulated_messages(  # type: ignore
        self,
    ) -> List[messages.CreateExperimentItemsBatchMessage]:
        return [
            messages.CreateExperimentItemsBatchMessage(
                batch=self._accumulated_messages,  # type: ignore
                supports_batching=False,
            )
        ]

    def add(  # type: ignore
        self, message: messages.CreateExperimentItemsBatchMessage
    ) -> None:
        new_messages = message.batch
        n_new_messages = len(new_messages)
        n_accumulated_messages = len(self._accumulated_messages)

        if n_new_messages + n_accumulated_messages >= self._max_batch_size:
            free_space_in_accumulator = self._max_batch_size - n_accumulated_messages

            messages_that_fit_in_batch = new_messages[:free_space_in_accumulator]
            messages_that_dont_fit_in_batch = new_messages[free_space_in_accumulator:]

            self._accumulated_messages += messages_that_fit_in_batch
            new_messages = messages_that_dont_fit_in_batch
            self.flush()

        self._accumulated_messages += new_messages



from typing import Any, List, Optional, Tuple, Union

from .parameter import Parameter


def validate_type_str(value: Any, allow_empty: bool = True) -> bool:
    valid = validate_type(value, (str,), allow_empty=allow_empty)
    if not valid:
        return False

    if not allow_empty:
        return value != ""

    return True


def validate_type_int(value: Any, allow_empty: bool = True) -> bool:
    return validate_type(value, (int,), allow_empty=allow_empty)


def validate_type_float(value: Any, allow_empty: bool = True) -> bool:
    return validate_type(value, (float,), allow_empty=allow_empty)


def validate_type_numeric(value: Any, allow_empty: bool = True) -> bool:
    return validate_type(value, (float, int), allow_empty=allow_empty)


def validate_type_bool(value: Any, allow_empty: bool = True) -> bool:
    return validate_type(value, (bool,), allow_empty=allow_empty)


def validate_type_list(value: Any, allow_empty: bool = True) -> bool:
    return validate_type(value, (list,), allow_empty=allow_empty)


def validate_type_list_of_numbers(value: List[Any]) -> bool:
    if not validate_type_list(value, allow_empty=False):
        return False
    res = [validate_type_numeric(n, allow_empty=False) for n in value]
    return all(res)


def validate_type_dict(value: Any, allow_empty: bool = True) -> bool:
    return validate_type(value, (dict,), allow_empty=allow_empty)


def validate_type(value: Any, types: Tuple, allow_empty: bool) -> bool:
    if value is None:
        return allow_empty

    return any(isinstance(value, t) for t in types)


def validate_possible_values(
    value: Any, possible_values: Optional[List], allow_empty: bool
) -> bool:
    if value is None:
        return allow_empty

    if possible_values is None:
        return False

    return value in possible_values


def validate_value(value: Any, allow_empty: bool) -> bool:
    if value is None:
        return allow_empty

    if not allow_empty and hasattr(value, "__len__") and len(value) == 0:
        return False

    return True


def validate_parameter(parameter: Parameter) -> Tuple[bool, Optional[str]]:
    if parameter.possible_values is not None:
        valid_value = validate_possible_values(
            value=parameter.value,
            possible_values=parameter.possible_values,
            allow_empty=parameter.allow_empty,
        )
    else:
        valid_value = validate_value(
            value=parameter.value, allow_empty=parameter.allow_empty
        )

    if not valid_value:
        if parameter.possible_values is not None:
            possible_values_str = [str(v) for v in parameter.possible_values]

            if parameter.allow_empty:
                msg = "parameter %r must be one of [%s] or None but %r was given" % (
                    parameter.name,
                    ", ".join(possible_values_str),
                    parameter.value,
                )
            else:
                msg = "parameter %r must be one of [%s] but %r was given" % (
                    parameter.name,
                    ", ".join(possible_values_str),
                    parameter.value,
                )
        else:
            msg = "parameter %r must have non empty value but %r was given" % (
                parameter.name,
                parameter.value,
            )

        return False, msg

    valid_type = validate_type(
        value=parameter.value, types=parameter.types, allow_empty=parameter.allow_empty
    )

    if not valid_type:
        param_type = None if parameter.value is None else type(parameter.value).__name__

        if parameter.allow_empty:
            msg = "parameter %r must be of type(s) %r or None but %r was given" % (
                parameter.name,
                types_list(parameter.types),
                param_type,
            )
        else:
            msg = "parameter %r must be of type(s) %r but %r was given" % (
                parameter.name,
                types_list(parameter.types),
                param_type,
            )
        return False, msg

    return True, None


def types_list(types: Union[Tuple, List]) -> Union[str, List[str]]:
    type_names = []
    for t in types:
        type_names.append(t.__name__)

    if len(type_names) > 1:
        return type_names
    elif len(type_names) == 1:
        return type_names[0]
    else:
        return []



import logging
from typing import Any, Optional, Type, List, Dict, TYPE_CHECKING
import pydantic

from . import opik_monitoring, message_converters
from ...models import base_model

if TYPE_CHECKING:
    import langchain_core.language_models
    import langchain_core.messages

LOGGER = logging.getLogger(__name__)


class LangchainChatModel(base_model.OpikBaseModel):
    def __init__(
        self,
        chat_model: "langchain_core.language_models.BaseChatModel",
        track: bool = True,
    ) -> None:
        """
        Initializes the model with a given Langchain chat model instance.

        Args:
            chat_model: A Langchain chat model instance to wrap.
                It is assumed that the BaseChatModel is already configured and
                all the requirement dependencies are installed.
            track: Whether to track the model calls.
        """
        model_name = _extract_model_name(chat_model)
        super().__init__(model_name=model_name)

        self._engine = chat_model
        self._track = track

    def generate_string(
        self,
        input: str,
        response_format: Optional[Type[pydantic.BaseModel]] = None,
        **kwargs: Any,
    ) -> str:
        """
        Simplified interface to generate a string output from the model.

        Args:
            input: The input string based on which the model will generate the output.
            response_format: pydantic model specifying the expected output string format.
            kwargs: Additional arguments that may be used by the model for string generation.

        Returns:
            str: The generated string output.
        """
        if response_format is not None:
            kwargs["response_format"] = response_format

        request = [
            {
                "content": input,
                "role": "user",
            },
        ]
        with base_model.get_provider_response(
            model_provider=self, messages=request, **kwargs
        ) as response:
            return base_model.check_model_output_string(response.content)

    def generate_provider_response(
        self,
        messages: List[Dict[str, Any]],
        **kwargs: Any,
    ) -> "langchain_core.messages.AIMessage":
        """
        Do not use this method directly. It is intended to be used within `base_model.get_provider_response()` method.

        Generate a provider-specific response using the Langchain model.

        Args:
            messages: A list of messages to be sent to the model, should be a list of dictionaries with the keys
                "content" and "role".
            kwargs: arguments required by the provider to generate a response.

        Returns:
            ModelResponse: The response from the model provider.
        """
        langchain_messages = message_converters.convert_to_langchain_messages(messages)

        opik_monitoring.add_opik_tracer_to_params(kwargs)
        response = self._engine.invoke(langchain_messages, **kwargs)

        return response

    async def agenerate_string(
        self,
        input: str,
        response_format: Optional[Type[pydantic.BaseModel]] = None,
        **kwargs: Any,
    ) -> str:
        """
        Simplified interface to generate a string output from the model. Async version.

        Args:
            input: The input string based on which the model will generate the output.
            response_format: pydantic model specifying the expected output string format.
            kwargs: Additional arguments that may be used by the model for string generation.

        Returns:
            str: The generated string output.
        """
        if response_format is not None:
            kwargs["response_format"] = response_format

        request = [
            {
                "content": input,
                "role": "user",
            },
        ]

        async with base_model.aget_provider_response(
            model_provider=self, messages=request, **kwargs
        ) as response:
            return base_model.check_model_output_string(response.content)

    async def agenerate_provider_response(
        self, messages: List[Dict[str, Any]], **kwargs: Any
    ) -> "langchain_core.messages.AIMessage":
        """
        Do not use this method directly. It is intended to be used within `base_model.aget_provider_response()` method.

        Generate a provider-specific response using the Langchain model. Async version.

        Args:
            messages: A list of messages to be sent to the model, should be a list of dictionaries with the keys
                "content" and "role".
            kwargs: arguments required by the provider to generate a response.

        Returns:
            ModelResponse: The response from the model provider.
        """
        langchain_messages = message_converters.convert_to_langchain_messages(messages)

        opik_monitoring.add_opik_tracer_to_params(kwargs)
        response = await self._engine.ainvoke(langchain_messages, **kwargs)

        return response


def _extract_model_name(
    langchain_chat_model: "langchain_core.language_models.BaseChatModel",
) -> str:
    if hasattr(langchain_chat_model, "model") and isinstance(
        langchain_chat_model.model, str
    ):
        return langchain_chat_model.model

    if hasattr(langchain_chat_model, "model_name") and isinstance(
        langchain_chat_model.model_name, str
    ):
        return langchain_chat_model.model_name

    if hasattr(langchain_chat_model, "model_id") and isinstance(
        langchain_chat_model.model_id, str
    ):
        return langchain_chat_model.model_id

    return "unknown-model"



import logging
from typing import Any, Dict, Optional

import opik
from opik import llm_usage, logging_messages, _logging as opik_logging
from . import langchain_run_helpers
from . import provider_usage_extractor_protocol
from .langchain_run_helpers import langchain_usage

LOGGER = logging.getLogger(__name__)


VERTEXAI_CANDIDATE_USAGE_KEYS = {
    "candidates_token_count",
    "prompt_token_count",
    "total_token_count",
}


class VertexAIUsageExtractor(
    provider_usage_extractor_protocol.ProviderUsageExtractorProtocol
):
    PROVIDER = opik.LLMProvider.GOOGLE_VERTEXAI

    def is_provider_run(self, run_dict: Dict[str, Any]) -> bool:
        try:
            if run_dict.get("serialized") is None:
                return False

            if (
                ls_metadata := langchain_run_helpers.try_get_ls_metadata(run_dict)
            ) is not None:
                if "google_vertexai" == ls_metadata.provider:
                    return True

            if (
                invocation_params := run_dict["extra"].get("invocation_params")
            ) is not None:
                if "vertexai" == invocation_params.get("_type"):
                    return True

            return False
        except Exception:
            LOGGER.debug(
                "Failed to check if Run instance is from VertexAI, returning False.",
                exc_info=True,
            )
            return False

    def get_llm_usage_info(
        self,
        run_dict: Dict[str, Any],
    ) -> llm_usage.LLMUsageInfo:
        usage_dict = _try_get_token_usage(run_dict)
        model = _try_get_model_name(run_dict)

        return llm_usage.LLMUsageInfo(
            provider=self.PROVIDER, model=model, usage=usage_dict
        )


def _try_get_token_usage(run_dict: Dict[str, Any]) -> Optional[llm_usage.OpikUsage]:
    try:
        if token_usage := langchain_run_helpers.try_to_get_usage_by_search(
            run_dict, VERTEXAI_CANDIDATE_USAGE_KEYS
        ):
            if isinstance(token_usage, dict):
                return llm_usage.OpikUsage.from_google_dict(token_usage)
            elif isinstance(token_usage, langchain_usage.LangChainUsage):
                gemini_usage_dict = token_usage.map_to_google_gemini_usage()
                return llm_usage.OpikUsage.from_google_dict(gemini_usage_dict)

        opik_logging.log_once_at_level(
            logging.WARNING,
            "Failed to extract token usage from presumably VertexAI LLM langchain run. Run dict: %s",
            LOGGER,
            run_dict,
        )

        opik_logging.log_once_at_level(
            logging_level=logging.WARNING,
            message=logging_messages.WARNING_TOKEN_USAGE_DATA_IS_NOT_AVAILABLE,
            logger=LOGGER,
        )
    except Exception:
        LOGGER.warning(
            logging_messages.FAILED_TO_EXTRACT_TOKEN_USAGE_FROM_PRESUMABLY_LANGCHAIN_GOOGLE_LLM_RUN,
            run_dict,
            exc_info=True,
        )

    return None


def _try_get_model_name(run_dict: Dict[str, Any]) -> Optional[str]:
    model = None

    # try metadata first
    if (ls_metadata := langchain_run_helpers.try_get_ls_metadata(run_dict)) is not None:
        model = ls_metadata.model

    elif (invocation_params := run_dict["extra"].get("invocation_params")) is not None:
        model = invocation_params.get("model_name")

    if model is not None:
        # Gemini **may** add "models/" prefix to some model versions
        model = model.split("/")[-1]

    return model
